import bisect

class SortedList(list):
    """
    Sorted list of objects that can keep iterating after one or more of its
    elements disappears, including the current one.
    """
    def __init__(self, initial=None):
        if initial is not None:
            if isinstance(initial, str):
                self.append(initial)
            else:
                self.extend(sorted(initial))

    def _index_of(self, value):
        pos = bisect.bisect_left(self, value)
        if pos != len(self) and self[pos] == value:
            return pos
        return None

    def add(self, value) -> None:
        if self._index_of(value) is None:
            bisect.insort(self, value)

    def remove(self, value) -> None:
        pos = self._index_of(value)
        if pos is not None:
            del self[pos]

    def first(self):
        return self[0] if len(self) > 0 else None

    def next(self, value):
        if len(self) == 0:
            return None
        pos = bisect.bisect(self, value)
        return None if pos == len(self) else self[pos]

    def last(self):
        return self[-1] if len(self) > 0 else None

    def prev(self, value):
        if len(self) == 0:
            return None
        pos = bisect.bisect_left(self, value)
        return None if pos == 0 else self[pos - 1]

    def has(self, value) -> bool:
        return self._index_of(value) is not None



class Routine:
    def __init__(self, name, address=0):
        self._name = name
        self._address = address
        self._params = []

    @property
    def name(self):
        return self._name

    @property
    def params(self):
        return self._params

    def add_param(self, name):
        self._params.append(name)

    def has_param(self, name):
        return name in self._params

    def set_address(self, address):
        self._address = address

    def get_address(self):
        return self._address



import functools

_providers = {}
injected = None


class UnboundException(Exception):
    def __init__(self, intf):
        super().__init__('No implementation for {}'.format(intf))


class Binder:
    def __init__(self, constructor):
        self._constructor = constructor

    def to(self, interface):
        _providers[interface] = self._constructor


class ObjectBinder:
    def __init__(self, instance):
        self._instance = instance

    def to(self, interface):
        _providers[interface] = lambda: self._instance


def inject(interface):
    def fn_wrapper(fn):
        @functools.wraps(fn)
        def param_wrapper(*args, **kwargs):
            return fn(*args, provide(interface), **kwargs)
        return param_wrapper
    return fn_wrapper


def provide(interface):
    if interface not in _providers:
        msg = "interface {}".format(interface)
        raise UnboundException(msg)
    return _providers[interface]()


def configure():
    global _providers
    _providers.clear()


def bind(implementation):
    return Binder(implementation)


def bind_instance(implementor):
    return ObjectBinder(implementor)



import os
import string

from bardolph.lib.symbol import SymbolType
from bardolph.parser.token import TokenTypes
from bardolph.vm.vm_codes import IoOp, OpCode, Register

from .sub_parser import SubParser


class IoParser(SubParser):
    def print(self) -> bool:
        self.next_token()
        if not self._out_current_token():
            return False
        self.code_gen.add_instruction(OpCode.OUT, IoOp.PRINT, ' ')
        return True

    def println(self) -> bool:
        self.next_token()
        if not self._out_current_token():
            return False
        self.code_gen.add_instruction(OpCode.OUT, IoOp.PRINT, os.linesep)
        return True

    def printf(self) -> bool:
        self.next_token()
        format_str = self.current_str
        if len(format_str) == 0:
            return self.token_error('Expected format specifier, got {}')
        self.next_token()
        for field in string.Formatter().parse(format_str):
            spec = field[1] or ''
            if len(spec) == 0 or spec.isdecimal():
                if not self._out_current_token():
                    return False
        self.code_gen.add_instruction(OpCode.OUT, IoOp.PRINTF, format_str)
        return True

    def _out_current_token(self) -> bool:
        if not self.rvalue():
            return False
        self.code_gen.add_instruction(
            OpCode.OUT, IoOp.REGISTER, Register.RESULT)
        return True



import logging
import string

from .vm_codes import IoOp, Register

class VmIo:
    def __init__(self, call_stack, reg):
        self._call_stack = call_stack
        self._reg = reg
        self._unnamed = []

    def out(self, inst):
        io_op = inst.param0
        if io_op is IoOp.REGISTER:
            self._unnamed.append(self._reg.get_by_enum(inst.param1))
        elif io_op is IoOp.PRINT:
            print(self._unnamed[0], end=inst.param1)
            self._unnamed.clear()
        elif io_op is IoOp.PRINTF:
            self._printf(inst)
        else:
            logging.error("Unrecognized output: {}".format(io_op))

    def reset(self):
        self._unnamed.clear()

    def _printf(self, inst):
        format_str = inst.param1
        params = {}
        for field in string.Formatter().parse(format_str):
            name = field[1] or ''
            if len(name) > 0 and not name.isdecimal():
                reg = Register.from_string(name)
                if reg is not None:
                    params[name] = self._reg.get_by_enum(reg)
                else:
                    params[name] = self._call_stack.get_variable(name)
        print(format_str.format(*self._unnamed, **params))
        self._unnamed.clear()



from datetime import datetime
import threading
import time

from . import injection
from . import i_lib

def now():
    # seconds
    return time.time()


def configure():
    injection.bind(Clock).to(i_lib.Clock)


# All time quantities are in seconds.
class Clock(i_lib.Clock):
    def __init__(self):
        self._event = threading.Event()
        self._start_time = 0.0
        self._cue_time = 0.0
        self._keep_going = True

    def start(self):
        self.reset()
        threading.Thread(target=self.run, args=(), daemon=True).start()

    @injection.inject(i_lib.Settings)
    def run(self, settings):
        self._keep_going = True
        sleep_time = float(settings.get_value('sleep_time'))
        while self._keep_going:
            if sleep_time > 0.0:
                time.sleep(sleep_time)
            self.fire()

    def stop(self):
        self._keep_going = False

    def reset(self):
        self._cue_time = 0.0
        self._start_time = now()

    def et(self):
        return time.time() - self._start_time

    def fire(self):
        self._event.set()
        self._event.clear()

    def wait(self):
        if self._keep_going:
            self._event.wait()
        return self._keep_going

    def pause_for(self, delay):
        self._cue_time += delay
        while self.et() < self._cue_time:
            if not self.wait():
                break

    def wait_until(self, time_pattern):
        hour, minute = Clock._hour_minute()
        while not time_pattern.match(hour, minute):
            self.wait()
            hour, minute = Clock._hour_minute()
        self.reset()

    @staticmethod
    def _hour_minute():
        now = datetime.now()
        return (now.hour, now.minute)



#!/usr/bin/env python

import argparse
from numbers import Number

from bardolph.lib import injection
from bardolph.lib import settings
from bardolph.vm.vm_codes import Register

from . import arg_helper
from . import config_values
from . import light_module
from .i_controller import LightSet
from . import lsc
from . import units


def _quote_if_string(param):
    return ('{}' if isinstance(param, str) else "{:0.0f}").format(param)

class Snapshot:
    def start_snapshot(self): pass
    def start_light(self, light): pass
    def record_setting(self, name, value): pass
    def handle_power(self, power): pass
    def end_light(self): pass
    def end_snapshot(self): pass
    def start_multizone(self, light): pass
    def handle_zone(self, light, number, color): pass
    def end_multizone(self, light): pass

    def handle_color(self, color):
        self.record_setting(Register.HUE, color[0])
        self.record_setting(Register.SATURATION, color[1])
        self.record_setting(Register.BRIGHTNESS, color[2])
        self.record_setting(Register.KELVIN, color[3])

    def handle_zones(self, light):
        for number, color in enumerate(light.get_color_zones()):
            self.handle_zone(light, number, color)

    @injection.inject(LightSet)
    def generate(self, light_set):
        self.start_snapshot()
        for name in light_set.light_names:
            light = light_set.get_light(name)
            if light.multizone:
                self.start_multizone(light)
                self.handle_zones(light)
                self.end_multizone(light)
            else:
                self.start_light(light)
                self.handle_color(light.get_color())
                self.handle_power(light.get_power())
                self.end_light()
        self.end_snapshot()
        return self


class ScriptSnapshot(Snapshot):
    """ Generate a .ls _script. """
    def __init__(self):
        self._light_name = ''
        self._power = True
        self._script = ''

    def start_snapshot(self):
        self._script = 'time 0 duration 1.5\n'

    def start_light(self, light):
        self._light_name = light.name

    def record_setting(self, reg, value):
        self._script += '{} {} '.format(reg, value)

    def handle_color(self, raw_color):
        # Output is always logical units.
        logical_color = units.raw_to_logical(raw_color)
        params = zip([
            Register.HUE, Register.SATURATION, Register.BRIGHTNESS,
            Register.KELVIN], logical_color)
        for param in params:
            reg, value = param[0], param[1]
            fmt = '{} {:.0f} ' if isinstance(value, Number) else '{} {} '
            self._script += fmt.format(reg.name.lower(), value)

    def handle_power(self, power):
        self._power = power

    def end_light(self):
        self._script += 'set "{}"\n'.format(self._light_name)
        fmt = 'on "{}"\n' if self._power else 'off "{}"\n'
        self._script += fmt.format(self._light_name)

    def start_multizone(self, light):
        self._light_name = light.name
        self._power = light.get_power()

    def handle_zone(self, light, number, color):
        self.handle_color(color)
        self._script += 'set "{}" zone {}\n'.format(light.name, number)

    @property
    def text(self):
        return '{}\n'.format(self._script)


class InstructionSnapshot(Snapshot):
    """ Generate a list of lists, one for each light. """
    def __init__(self):
        self._light_name = ''
        self._snapshot = ''
        self._power = ''

    def start_snapshot(self):
        self._snapshot = ''

    def start_light(self, light):
        self._light_name = light.name

    def record_setting(self, reg, value):
        self._snapshot += '    OpCode.MOVEQ, {}, {},\n'.format(
            _quote_if_string(value), reg)

    def handle_power(self, power):
        self._power = power

    def end_light(self):
        self._snapshot += '    OpCode.MOVEQ, "{}", Register.NAME,\n'.format(
            self._light_name)
        self._snapshot += '    OpCode.MOVEQ, Operand.LIGHT, Register.OPERAND,\n'
        self._snapshot += '    OpCode.COLOR,\n'
        self._snapshot += '    OpCode.MOVEQ, {}, Register.POWER,\n'.format(
                            self._power)
        self._snapshot += '    OpCode.POWER,\n'

    @property
    def text(self):
        return self._snapshot[:-2]


class TextSnapshot(Snapshot):
    """ Generate plain text. """
    def __init__(self):
        self._field_width = 10
        self._text = ''
        self._add_field('name ')._add_field(' hue')
        self._add_field(' sat')._add_field(' brt')
        self._add_field(' kel')._add_field('power')
        self._text += '\n'
        self._text += '-' * ((self._field_width) * 6 - 5)
        self._text += '\n'

    def _add_field(self, data):
        self._text += str(data).ljust(self._field_width)
        return self

    @injection.inject(LightSet)
    def _add_sets(self, lights):
        self._add_set('Groups', lights.group_names, lights.get_group)
        self._add_set(
            'Locations', lights.location_names, lights.get_location)

    def _add_set(self, heading, names, get_fn):
        self._text += '\n{}\n'.format(heading)
        self._text += '-' * 15
        self._text += '\n'
        for name in names:
            self._text += '{}\n'.format(name)
            for light in get_fn(name):
                self._text += '   {}\n'.format(light)

    def generate(self):
        super().generate()
        self._add_sets()
        return self

    def start_light(self, light):
        self._add_field(light.name)

    def start_multizone(self, light):
        self.start_light(light)
        self._text += '{:>45}'.format(light.get_power())
        self._text += '\n   Zone\n'

    def handle_zone(self, _, number, color):
        self._add_field('{:>5d}'.format(number))
        self.handle_color(color)
        self._text += '\n'

    def end_zones(self, _):
        self._text += '\n'

    def handle_color(self, raw_color):
        logical_color = units.raw_to_logical(raw_color)
        for value in logical_color:
            self._add_field('{:>4.0f}'.format(value))

    def handle_power(self, power):
        self._add_field('{:d}'.format(power))

    def end_light(self):
        self._text += '\n'

    @property
    def text(self):
        return self._text


class DictSnapshot(Snapshot):
    """ Generate a list of dictionaries, one for each light. """
    def __init__(self):
        self._snapshot = None
        self._current_dict = None

    def start_snapshot(self):
        self._snapshot = []
        self._current_dict = {}

    def start_light(self, light):
        self._current_dict = {'_name': light.name}

    def record_setting(self, name, value):
        self._current_dict[name] = value

    def end_light(self):
        self._snapshot.append(self._current_dict)

    @property
    def snapshot(self):
        return self._snapshot

    @property
    def text(self):
        return str(self._current_dict)


def _do_gen(ctor):
    print(ctor().generate().text + '\n')


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-l', '--list', help='output instruction list', action='store_true')
    parser.add_argument(
        '-d', '--dict', help='output dictionary format', action='store_true')
    parser.add_argument(
        '-f', '--use-fakes', help='use fake lights', action='store_true')
    arg_helper.add_n_argument(parser)
    parser.add_argument(
        '-p', '--py', help='output Python code', action='store_true')
    parser.add_argument(
        '-s', '--script', help='output script format', action='store_true')
    parser.add_argument(
        '-t', '--text', help='output text format', action='store_true')
    args = parser.parse_args()
    do_script = args.script
    do_dict = args.dict
    do_list = args.list
    do_py = args.py
    do_text = args.text or (not (do_py or do_script or do_dict or do_list))

    injection.configure()
    settings_init = settings.using(
        config_values.functional).add_overrides({'single_light_discover': True})
    settings_init.apply_env()

    if args.use_fakes:
        settings_init.add_overrides({'use_fakes': True})
    n_arg = arg_helper.get_overrides(args)
    if n_arg is not None:
        settings_init.add_overrides(n_arg)
    settings_init.configure()
    light_module.configure()

    if do_dict:
        _do_gen(DictSnapshot)
    if do_list:
        _do_gen(InstructionSnapshot)
    if do_script:
        _do_gen(ScriptSnapshot)
    if do_text:
        _do_gen(TextSnapshot)
    if do_py:
        snap = InstructionSnapshot()
        text = '    OpCode.MOVEQ, UnitMode.RAW, Register.UNIT_MODE,\n'
        text += snap.generate().text
        lsc.output_python(lsc.program_code(text))


if __name__ == '__main__':
    main()



import logging

import lifxlan

from . import i_controller
from ..lib import i_lib
from ..lib.injection import bind, inject

class Lifx(i_controller.Lifx):
    @inject(i_lib.Settings)
    def __init__(self, settings):
        num_expected = settings.get_value('default_num_lights', None)
        self._lifxlan = lifxlan.LifxLAN(num_expected)
    
    @inject(i_lib.Settings)
    def get_lights(self, settings):
        lights = self._lifxlan.get_lights()
        expected = settings.get_value('default_num_lights', None)
        if expected is not None:
            actual = len(lights)
            if actual < expected:
                logging.warning(
                    "Expected {} devices, found {}".format(expected, actual))
        return lights
    
    def set_color_all_lights(self, color, duration):
        self._lifxlan.set_color_all_lights(color, duration, True)
    
    def set_power_all_lights(self, power_level, duration):
        self._lifxlan.set_power_all_lights(power_level, duration, True)


def configure():
    bind(Lifx).to(i_controller.Lifx)



#!/usr/bin/env python

import argparse
import logging

from bardolph.lib import injection
from bardolph.lib import settings
from bardolph.controller import arg_helper
from bardolph.controller import config_values
from bardolph.controller import light_module
from bardolph.controller.units import UnitMode
from bardolph.vm import machine
from bardolph.vm.instruction import Instruction, OpCode
from bardolph.vm.vm_codes import IoOp, JumpCondition, LoopVar, Operand, Operator
from bardolph.vm.vm_codes import Register, SetOp

_assembly = [
    #instructions

]

_param_counts = {op_code: 0 for op_code in (OpCode.BREAKPOINT, OpCode.COLOR,
    OpCode.DISC, OpCode.END_LOOP, OpCode.GET_COLOR, OpCode.LOOP, OpCode.NOP,
    OpCode.PAUSE, OpCode.STOP, OpCode.POWER, OpCode.WAIT)}
_param_counts.update({op_code: 1 for op_code in (OpCode.DISCM, OpCode.DNEXT,
    OpCode.END, OpCode.JSR, OpCode.OP, OpCode.POP, OpCode.PUSH, OpCode.PUSHQ,
    OpCode.ROUTINE)})
_param_counts.update({op_code: 2 for op_code in (OpCode.CONSTANT, OpCode.DNEXTM,
    OpCode.JUMP, OpCode.MOVE, OpCode.MOVEQ, OpCode.OUT, OpCode.PARAM,
    OpCode.TIME_PATTERN)})

def get_assembly():
    current_instruction = 0
    while current_instruction < len(_assembly):
        value = _assembly[current_instruction]
        current_instruction += 1
        yield value

def build_instructions():
    program = []
    it = iter(get_assembly())
    op_code = next(it, None)
    while op_code is not None:
        param_count = _param_counts[OpCode(op_code)]
        param0 = None if param_count < 1 else next(it)
        param1 = None if param_count < 2 else next(it)
        program.append(Instruction(op_code, param0, param1))
        op_code = next(it, None)
    return program

def main():
    injection.configure()

    ap = argparse.ArgumentParser()
    ap.add_argument(
        '-v', '--verbose', help='do debug-level logging', action='store_true')
    ap.add_argument(
        '-f', '--fakes', help='use fake lights', action='store_true')
    arg_helper.add_n_argument(ap)
    args = ap.parse_args()

    overrides = {
        'sleep_time': 0.1
    }
    if args.verbose:
        overrides['log_level'] = logging.DEBUG
        overrides['log_to_console'] = True
    if args.fakes:
        overrides['use_fakes'] = True
    n_arg = arg_helper.get_overrides(args)
    if n_arg is not None and not args.fakes:
        overrides.update(n_arg)

    settings_init = settings.using(config_values.functional)
    settings_init.add_overrides(overrides).configure()
    light_module.configure()
    machine.Machine().run(build_instructions())


if __name__ == '__main__':
    main()



from bardolph.vm.vm_codes import OpCode, Operator, Register
from bardolph.parser.token import Assoc, TokenTypes
from bardolph.parser.sub_parser import SubParser


class ExpressionParser(SubParser):
    def expression(self):
        return self._atom() and self._expression(0)

    def _expression(self, min_prec) -> bool:
        while (self.current_token.is_binop
                and self.current_token.prec >= min_prec):
            op = self.current_token
            self.next_token()
            if not self._atom():
                return False
            while ((self.current_token.is_binop
                        and self.current_token.prec > op.prec)
                    or (self.current_token.assoc is Assoc.RIGHT
                        and self.current_token.prec == op.prec)):
                if not self._expression(self.current_token.prec):
                    return False
            if not self._do_op(op):
                return False
        return True

    def _atom(self) -> bool:
        if str(self.current_token) == '(':
            self.next_token()
            if not self.expression():
                return False
            if self.current_token != ')':
                return self.token_error('Unmatched parenthesis: {}')
            return self.next_token()

        uminus = str(self.current_token) == '-'
        if str(self.current_token) in '+-':
            self.next_token()
            if not self._atom():
                return False
            if uminus:
                self.code_gen.add_list(
                    (OpCode.PUSHQ, -1),
                    (OpCode.OP, Operator.MUL)
                )
        elif not self.rvalue(OpCode.PUSH):
            return False
        return True

    def _do_op(self, op) -> bool:
        # Each of these will pop two arguments off the stack, perform the
        # calculation, and push the result.
        operator = {
            '+': Operator.ADD,
            '-': Operator.SUB,
            '*': Operator.MUL,
            '/': Operator.DIV,
            '^': Operator.POW,
            'and': Operator.AND,
            'or': Operator.OR,
            '<': Operator.LT,
            '<=': Operator.LTE,
            '>': Operator.GT,
            '>=': Operator.GTE,
            '==': Operator.EQ,
            '!=': Operator.NOTEQ}.get(op.content)
        if operator is None:
            return self.token_error('Invalid operand {} in expression.')
        self.code_gen.add_instruction(OpCode.OP, operator)
        return True



from bardolph.lib import clock
from bardolph.lib import log_config
from bardolph.lib.i_lib import Settings
from bardolph.lib.injection import provide

from . import light_set

def configure():
    """ Assumes injection and settings are already initialized. """
    log_config.configure()
    clock.configure()

    settings = provide(Settings)
    if settings.get_value('use_fakes'):
        from ..fakes import fake_lifx
        fake_lifx.configure()
    else:
        from . import lifx
        lifx.configure()

    light_set.configure()



from .eval_stack import EvalStack
from .vm_codes import LoopVar, Operand, Operator, Register

from bardolph.controller.i_controller import LightSet
from bardolph.lib.injection import inject, injected, provide
from bardolph.lib.sorted_list import SortedList

class VmDiscover:
    def __init__(self, call_stack, reg):
        self._call_stack = call_stack
        self._reg = reg

    def disc(self) -> None:
        name_list = self._names_by_oper()
        if len(name_list) == 0:
            self._reg.result = Operand.NULL
        else:
            index = 0 if self._reg.disc_forward else -1
            self._reg.result = name_list[index]

    def discm(self, name) -> None:
        name_list = self._set_by_oper(self._param_to_value(name))
        if name_list and len(name_list) > 0:
            index = 0 if self._reg.disc_forward else -1
            self._reg.result = name_list[index] or Operand.NULL
        else:
            self._reg.result = Operand.NULL

    def dnext(self, current) -> None:
        name_list = self._names_by_oper()
        current = self._param_to_value(current)
        if not self._reg.disc_forward:
            self._reg.result = name_list.prev(current) or Operand.NULL
        else:
            self._reg.result = name_list.next(current) or Operand.NULL

    def dnextm(self, name, current) -> None:
        name_list = self._set_by_oper(self._param_to_value(name))
        current = self._param_to_value(current)
        if not self._reg.disc_forward:
            self._reg.result = name_list.prev(current) or Operand.NULL
        else:
            self._reg.result = name_list.next(current) or Operand.NULL

    def _param_to_value(self, param):
        if isinstance(param, (str, Operand)):
            return param
        if isinstance(param, Register):
            return self._reg.get_by_enum(param)
        return self._call_stack.get_variable(param)

    @inject(LightSet)
    def _set_by_oper(self, name, light_set):
        if self._reg.operand is Operand.GROUP:
            return light_set.get_group(name)
        elif self._reg.operand is Operand.LOCATION:
            return light_set.get_location(name)
        return None

    @inject(LightSet)
    def _names_by_oper(self, light_set):
        if self._reg.operand is Operand.GROUP:
            return light_set.group_names
        elif self._reg.operand is Operand.LOCATION:
            return light_set.location_names
        assert self._reg.operand is Operand.LIGHT
        return light_set.light_names



import functools
import logging


def tries(num_tries, ex_type, fail_value=False):
    '''Decorator for retrying a function if a specific type of Exception is
    raised.

    Parameters
    ----------
    num_tries
        Total number of times to try before giving up. Should be >= 2.
    ex_type
        Type of exception that allows a retry. Any other type of exception
        raised by the wrapped function will be unhandled.
    fail_value:
        What value to return after all the tries have been exhausted. Defaults
        to False.

    Returns
    -------
        Whatever the wrapped function returns, or fail_value if the maximum
        number of tries has ocurred.
    '''
    def fn_wrapper(fn):
        @functools.wraps(fn)
        def param_wrapper(*args, **kwargs):
            tries_remaining = num_tries
            while tries_remaining > 0:
                try:
                    return fn(*args, **kwargs)
                except ex_type as ex:
                    logging.warning(ex)
                    tries_remaining -= 1
            logging.warning('Giving up after {} tries.'.format(num_tries))
            return fail_value
        return param_wrapper
    return fn_wrapper



import colorsys
from enum import Enum, auto

from bardolph.vm.vm_codes import Register

_EPSILON = 1.0 / 65536.0 / 2.0

class UnitMode(Enum):
    LOGICAL = auto()
    RAW = auto()
    RGB = auto()

def time_raw(logical_time):
    return logical_time * 1000.0

def time_logical(raw_time):
    raw_time = float(raw_time)
    return 0.0 if -_EPSILON < raw_time < _EPSILON else raw_time / 1000.0

def _pct_to_raw(pct):
    return 0.0 if -_EPSILON < pct < _EPSILON else pct / 100.0 * 65535.0

def logical_to_raw(logical_color):
    logical_value = logical_color[0]
    if (-_EPSILON < logical_value < _EPSILON or
            360 - _EPSILON < logical_value < 360 + _EPSILON):
        h = 0.0
    else:
        h = (logical_value % 360.0) / 360.0 * 65535.0
    s = _pct_to_raw(logical_color[1])
    b = _pct_to_raw(logical_color[2])
    return [h, s, b, logical_color[3]]

def raw_to_logical(raw_color):
    raw_value = raw_color[0]
    h = float(raw_value) / 65535.0 * 360.0
    raw_value = raw_color[1]
    s = 100.0 if raw_value >= 65535.0 else float(raw_value) / 65535.0 * 100.0
    raw_value = raw_color[2]
    b = 100.0 if raw_value >= 65535.0 else float(raw_value) / 65535.0 * 100.0
    return [h, s, b, raw_color[3]]

def rgb_to_raw(rgb_color):
    r, g, b = [rgb_color[i] / 100.0 for i in range(0, 3)]
    h, s, v = colorsys.rgb_to_hsv(r, g, b)
    make_raw = lambda x: min(round(x * 65535.0), 65535)
    return [make_raw(h), make_raw(s), make_raw(v), rgb_color[3]]

def rgb_to_logical(rgb_color):
    r, g, b = [rgb_color[i] / 100.0 for i in range(0, 3)]
    h, s, v = colorsys.rgb_to_hsv(r, g, b)
    return [h * 360.0, s * 100.0, v * 100.0, rgb_color[3]]

def raw_to_rgb(raw_color):
    h, s, v = [raw_color[i] / 65535.0 for i in range(0, 3)]
    r, g, b = colorsys.hsv_to_rgb(h, s, v)
    return [r * 100.0, g * 100.0, b * 100.0, raw_color[3]]

def logical_to_rgb(logical_color):
    h = logical_color[0] / 360.0
    s = logical_color[1] / 100.0
    v = logical_color[2] / 100.0
    r, g, b = colorsys.hsv_to_rgb(h, s, v)
    return [r * 100.0, g * 100.0, b * 100.0, logical_color[3]]



from collections import deque

class EvalStack:
    def __init__(self):
        self._stack = deque()

    def clear(self):
        self._stack.clear()

    @property
    def top(self):
        return self.below(0)

    def below(self, depth):
        return self._stack[-(depth + 1)]

    def push(self, value):
        self._stack.append(value)

    def pop(self):
        assert len(self._stack) > 0
        return self._stack.pop()

    def replace_top(self, value):
        self._stack[-1] = value



from .symbol import Symbol, SymbolType

class SymbolTable:
    def __init__(self):
        self._dict = {}

    def __contains__(self, name):
        return name in self._dict

    def clear(self):
        self._dict.clear()

    def add_symbol(self, name, symbol_type=SymbolType.UNKNOWN, value=None):
        self._dict[name] = Symbol(name, symbol_type, value)

    def get_symbol(self, name):
        return self._dict.get(name, Symbol())

    def get_type(self, name):
        return self.get_symbol(name).symbol_type

    def get_value(self, name):
        symbol = self.get_symbol(name)
        return None if symbol is None else symbol.value

    def get_routine(self, name):
        s_type = self.get_type(name)
        return self.get_value(name) if s_type == SymbolType.ROUTINE else None



from bardolph.controller.units import UnitMode
from bardolph.vm.instruction import Instruction
from bardolph.vm.vm_codes import JumpCondition, LoopVar, OpCode, Operand
from bardolph.vm.vm_codes import Operator, Register

class _JumpMarker:
    def __init__(self, inst, offset):
        self.jump = inst
        self.offset = offset


class CodeGen:
    def __init__(self):
        self._code = []

    @property
    def program(self):
        return self._code

    @property
    def current_offset(self) -> int:
        return len(self._code)

    def clear(self) -> None:
        self._code.clear()

    def push(self, operand) -> None:
        self.add_instruction(self._push_op(operand), operand)

    def pop(self, operand) -> None:
        self.add_instruction(OpCode.POP, operand)

    def add_instruction(self, op_code, param0=None, param1=None) -> Instruction:
        inst = Instruction(op_code, param0, param1)
        self._code.append(inst)
        return inst

    def add_list(self, *inst_list) -> None:
        # Convert a list of tuples to Instructions.
        for code in inst_list:
            if isinstance(code, OpCode):
                self.add_instruction(code)
            else:
                op_code, param0, param1, *_ = (*code, None, None)
                self.add_instruction(op_code, param0, param1)

    def add_instructions(self, inst_list) -> None:
        self._code.extend(inst_list)

    def add(self, addend0, addend1) -> None:
        self.binop(Operator.ADD, addend0, addend1)

    def subtract(self, minuend, subtrahend) -> None:
        """
        minuend - subtrahend
        Leaves the difference on top of the stack.
        """
        self.binop(Operator.SUB, minuend, subtrahend)

    def test_op(self, operator, op0, op1) -> None:
        """
        Generate code to perform a binary operation and put the results into
        the result register.
        """
        self.binop(operator, op0, op1)
        self.add_instruction(OpCode.POP, Register.RESULT)

    def binop(self, operator, param0, param1) -> None:
        push0 = self._push_op(param0)
        push1 = self._push_op(param1)
        self.add_list(
            (push0, param0),
            (push1, param1),
            (OpCode.OP, operator)
        )

    def plus_equals(self, dest, delta=1) -> None:
        self._op_equals(Operator.ADD, dest, delta)

    def minus_equals(self, dest, delta=1) -> None:
        self._op_equals(Operator.SUB, dest, delta)

    def times_equals(self, dest, pi) -> None:
        self._op_equals(Operator.MUL, dest, pi)

    def _op_equals(self, operator, original, change) -> None:
        push0 = self._push_op(original)
        push1 = self._push_op(change)
        self.add_list(
            (push0, original),
            (push1, change),
            (OpCode.OP, operator),
            (OpCode.POP, original)
        )

    def push_context(self, params) -> None:
        self.add_instruction(OpCode.JSR, params)

    def mark(self) -> _JumpMarker:
        return _JumpMarker(None, self.current_offset)

    def jump_back(self, marker) -> None:
        offset = marker.offset - self.current_offset
        self.add_instruction(OpCode.JUMP, JumpCondition.ALWAYS, offset)

    def if_true_start(self) -> _JumpMarker:
        inst = self.add_instruction(OpCode.JUMP, JumpCondition.IF_FALSE)
        return _JumpMarker(inst, self.current_offset)

    def if_else(self, marker) -> None:
        marker.jump.param1 = self.current_offset - marker.offset + 2
        inst = self.add_instruction(OpCode.JUMP, JumpCondition.ALWAYS)
        marker.jump = inst
        marker.offset = self.current_offset

    def if_end(self, marker) -> None:
        marker.jump.param1 = self.current_offset - marker.offset + 1

    def iter_lights(self, code) -> None:
        """ Generate code to iterate over all of the lights.
            moveq  operand.light   register.operand
            disc
        mark marker
            move register.result loopvar.current
            test operator.noteq loopvar.current operand.null
            if start
               ...insert code...
               moveq operand.light register.operand
               dnext loopvar.current
               jump mark
            if end
        """
        self.add_list(
            (OpCode.MOVEQ, Operand.LIGHT, Register.OPERAND),
            OpCode.DISC
        )
        loop_marker = self.mark()
        self.add_instruction(OpCode.MOVE, Register.RESULT, LoopVar.CURRENT)
        self.test_op(Operator.NOTEQ, LoopVar.CURRENT, Operand.NULL)
        if_marker = self.if_true_start()
        self._code.extend(list(code))
        self.add_list(
            (OpCode.MOVEQ, Operand.LIGHT, Register.OPERAND),
            (OpCode.DNEXT, LoopVar.CURRENT)
        )
        self.jump_back(loop_marker)
        self.if_end(if_marker)

    def iter_sets(self, operand, code) -> None:
        """ Iterate over all groups or locations. """
        self.add_instruction(OpCode.MOVEQ, operand, Register.OPERAND)
        self.add_instruction(OpCode.DISC)
        loop_marker = self.mark()
        self.add_instruction(OpCode.MOVE, Register.RESULT, LoopVar.CURRENT)
        self.test_op(Operator.NOTEQ, Register.RESULT, Operand.NULL)
        if_marker = self.if_true_start()
        self._code.extend(list(code))
        self.add_list(
            (OpCode.MOVEQ, operand, Register.OPERAND),
            (OpCode.DNEXT, LoopVar.CURRENT)
        )
        self.jump_back(loop_marker)
        self.if_end(if_marker)

    def iter_members(self, operand, code):
        """
        Generate code to iterate over all members of a group or location. The
        name must be in LoopVar.FIRST.
        """
        self.add_list(
            (OpCode.MOVEQ, operand, Register.OPERAND),
            (OpCode.DISCM, LoopVar.FIRST)
        )
        loop_marker = self.mark()
        self.add_instruction(OpCode.MOVE, Register.RESULT, LoopVar.CURRENT)
        self.test_op(Operator.NOTEQ, LoopVar.CURRENT, Operand.NULL)
        if_marker = self.if_true_start()
        self._code.extend(list(code))
        self.add_list(
            (OpCode.MOVEQ, operand, Register.OPERAND),
            (OpCode.DNEXTM, LoopVar.FIRST, LoopVar.CURRENT)
        )
        self.jump_back(loop_marker)
        self.if_end(if_marker)

    @staticmethod
    def _push_op(oper) -> OpCode:
        if isinstance(oper, (int, float, UnitMode)):
            return OpCode.PUSHQ
        return OpCode.PUSH



import logging
import time

from lifxlan.errors import WorkflowException

from bardolph.lib.color import rounded_color
from bardolph.lib.retry import tries


_MAX_TRIES = 3


class LightException(Exception):
    def __init__(self, cause):
        super().__init__('Exception from light API {}'.format(cause))
        self._cause = cause

    @property
    def cause(self):
        return self._cause


class Light:
    def __init__(self, lifx_light):
        self._impl = lifx_light
        self._name = lifx_light.get_label()
        self._group = lifx_light.get_group()
        self._location = lifx_light.get_location()
        self._multizone = lifx_light.supports_multizone()
        self._birth = time.time()

    def __repr__(self):
        fmt = 'Light(_name="{}", _group="{}", _location="{}", _multizone={}, '
        fmt += ' _birth={})'
        rep = fmt.format(
            self._name, self._group, self._location, self._multizone,
            self._birth)
        return rep

    @property
    def group(self):
        return self._group

    @property
    def location(self):
        return self._location

    @property
    def name(self):
        return self._name

    @property
    def multizone(self):
        return self._multizone

    def get_age(self):
        #seconds
        return time.time() - self._birth

    def set_color(self, color, duration, rapid=True):
        try:
            self._impl.set_color(rounded_color(color), duration, rapid)
        except WorkflowException as ex:
            logging.warning(ex)

    def get_color(self):
        try:
            return self._impl.get_color()
        except WorkflowException as ex:
            logging.warning(ex)
        return [-1] * 4

    @tries(_MAX_TRIES, WorkflowException)
    def set_zone_color(self, first_zone, last_zone, color, duration) -> None:
        # Unknown why this happens.
        if not hasattr(self._impl, 'set_zone_color'):
            logging.error(
                'No set_zone_color for light of type', type(self._impl))
        else:
            self._impl.set_zone_color(
                first_zone, last_zone, rounded_color(color), duration)

    @tries(_MAX_TRIES, WorkflowException)
    def get_color_zones(self, first_zone=None, last_zone=None):
        return self._impl.get_color_zones(first_zone, last_zone)

    @tries(_MAX_TRIES, WorkflowException)
    def set_power(self, power, duration, rapid=True):
        return self._impl.set_power(round(power), duration, rapid)

    @tries(_MAX_TRIES, WorkflowException)
    def get_power(self):
        return self._impl.get_power()



import logging

from bardolph.lib.job_control import Job
from bardolph.vm.machine import Machine, MachineState, Registers
from bardolph.parser.parse import Parser


class ScriptJob(Job):
    def __init__(self):
        super().__init__()
        self._program = None
        self._parser = Parser()
        self._machine = Machine()

    @staticmethod
    def from_file(file_name):
        new_instance = ScriptJob()
        new_instance.load_file(file_name)
        return new_instance

    @staticmethod
    def from_string(script):
        new_instance = ScriptJob()
        new_instance.load_string(script)
        return new_instance

    def load_file(self, file_name):
        self._program = self._parser.load(file_name)
        if self._program is None:
            logging.error(
                "{}, {}".format(file_name, self._parser.get_errors()))
        return self._program

    def load_string(self, input_string):
        self._program = self._parser.parse(input_string)
        if self._program is None:
            logging.error(self._parser.get_errors())
        return self._program

    @property
    def program(self):
        return self._program

    @property
    def compile_errors(self):
        return self._parser.get_errors()

    def get_machine_state(self) -> MachineState:
        return self._machine.get_state()

    def execute(self):
        if self._program is not None:
            self._machine.reset()
            self._machine.run(self._program)

    def request_stop(self):
        self._machine.stop()



#!/usr/bin/env python

import argparse
import logging
import os

from ..lib import injection
from ..lib import job_control
from ..lib import settings

from . import arg_helper
from . import light_module
from . import config_values
from .script_job import ScriptJob

_epilog = """The -n parameter is optional, but if you don't specify it,
discovery of the lights will take several seconds, and there will
be a noticeable pause before the script actually runs. If specified, this
parameter overrides any values in any configuration files."""


def init_args():
    parser = argparse.ArgumentParser(epilog=_epilog)
    parser.add_argument('file', help='name of the script file', nargs='*')
    parser.add_argument(
        '-c', '--config-file', help='customized configuration file')
    parser.add_argument(
        '-f', '--fakes', help='use fake lights', action='store_true')
    arg_helper.add_n_argument(parser)
    parser.add_argument(
        '-s', '--script', help='run script from command line', action='store')
    parser.add_argument(
        '-v', '--verbose', help='verbose output', action='store_true')
    return parser.parse_args()


def init_settings(args):
    overrides = {
        'log_date_format': "%I:%M:%S %p",
        'log_format': '%(asctime)s %(filename)s(%(lineno)d): %(message)s',
        'log_level': logging.DEBUG if args.verbose else logging.INFO,
        'log_to_console': True,
        'sleep_time': 0.1
    }
    if args.fakes:
        overrides['use_fakes'] = True

    settings_init = settings.using(
        config_values.functional).add_overrides(overrides)

    if args.config_file:
        settings_init.apply_file(args.config_file)

    overrides = arg_helper.get_overrides(args)
    if overrides is not None:
        settings_init.add_overrides(overrides)
    settings_init.apply_env().configure()


def main():
    args = init_args()
    injection.configure()
    init_settings(args)
    light_module.configure()

    jobs = job_control.JobControl()
    if args.script is not None:
        jobs.add_job(ScriptJob.from_string(args.script))
    for file_name in args.file:
        jobs.add_job(ScriptJob.from_file(file_name))


if __name__ == "__main__":
    main()



from enum import Enum, auto

from bardolph.controller.units import UnitMode
from bardolph.vm.vm_codes import LoopVar, OpCode, Operand, Operator, Register

from .code_gen import CodeGen
from .sub_parser import SubParser
from .token import TokenTypes


class _LoopType(Enum):
    ALL = auto()
    COUNTED = auto()
    GROUPS = auto()
    INFINITE = auto()
    LIST = auto()
    LOCATIONS = auto()
    WHILE = auto()
    WITH = auto()

    def is_unbounded(self):
        # Number of iterations is indefinite.
        return self in (_LoopType.INFINITE, _LoopType.WHILE)

    def is_iter(self):
        # Iteration that pushes names onto the stack.
        return self in (_LoopType.ALL, _LoopType.GROUPS, _LoopType.LIST,
            _LoopType.LOCATIONS)


class LoopParser(SubParser):
    def __init__(self, parser):
        super().__init__(parser)
        self._loop_type = None
        self._index_var = None
        self._light_var = None
        self._counter = None

    def repeat(self, code_gen, context_stack) -> bool:
        self.next_token()
        context_stack.enter_loop()
        code_gen.add_instruction(OpCode.LOOP)
        if not self._detect_loop_type():
            return False
        if not self._pre_loop(code_gen, context_stack):
            return False
        loop_top = code_gen.mark()
        if not self._loop_test(code_gen):
            return False
        exit_loop_marker = code_gen.if_true_start()
        if not (self._loop_body(code_gen) and self._loop_post(code_gen)):
            return False
        code_gen.jump_back(loop_top)
        code_gen.if_end(exit_loop_marker)
        context_stack.fix_break_addrs(code_gen)
        code_gen.add_instruction(OpCode.END_LOOP)
        context_stack.exit_loop()
        return True

    def _detect_loop_type(self) -> bool:
        self._loop_type = {
            TokenTypes.WHILE: _LoopType.WHILE,
            TokenTypes.WITH: _LoopType.WITH,
            TokenTypes.IN: _LoopType.LIST,
            TokenTypes.ALL: _LoopType.ALL,
            TokenTypes.GROUP: _LoopType.GROUPS,
            TokenTypes.LOCATION: _LoopType.LOCATIONS
        }.get(self.current_token.token_type)
        if self._loop_type is None:
            if self.at_rvalue():
                self._loop_type = _LoopType.COUNTED
            else:
                self._loop_type = _LoopType.INFINITE
        elif self._loop_type not in (_LoopType.ALL, _LoopType.WITH):
            self.next_token()
        return True

    def _pre_loop(self, code_gen, context_stack) -> bool:
        if self._loop_type.is_unbounded():
            return True

        if (self._loop_type is _LoopType.COUNTED
                and not self.rvalue(LoopVar.COUNTER)):
            return False
        if self._loop_type in (_LoopType.ALL, _LoopType.LIST):
            code_gen.add_instruction(OpCode.MOVEQ, 0, LoopVar.COUNTER)
            if not (self._pre_loop_list(code_gen, context_stack) and
                    self._pre_loop_as(context_stack)):
                return False
        elif self._loop_type in (_LoopType.GROUPS, _LoopType.LOCATIONS):
            code_gen.add_instruction(OpCode.MOVEQ, 0, LoopVar.COUNTER)
            if not (self._push_set_names(code_gen) and
                    self._pre_loop_as(context_stack)):
                return False

        if self.current_token.is_a(TokenTypes.WITH):
            self.next_token()
            return self._pre_loop_with(code_gen, context_stack)
        return True

    def _pre_loop_list(self, code_gen, context_stack) -> bool:
        """ Push everything from the "in" clause onto the stack. """
        if self.current_token.is_a(TokenTypes.AS):
            return True
        inner_coder = CodeGen()
        operand = {
            TokenTypes.ALL: Operand.LIGHT,
            TokenTypes.GROUP: Operand.GROUP,
            TokenTypes.LOCATION: Operand.LOCATION
        }.get(self.current_token.token_type)
        if operand is not None:
            # Push all lights, or all members of a group/location.
            self.next_token()
            if not self._push_light_names(inner_coder, operand):
                return False
        else:
            # Push and count one light.
            if not self.rvalue(Register.RESULT, inner_coder):
               return False
            inner_coder.push(Register.RESULT)
            inner_coder.plus_equals(LoopVar.COUNTER)

        if self.current_token.is_a(TokenTypes.AND):
            if operand == Operand.ALL:
                return self.trigger_error('"and" is not allowed with "all"')
            if not (self._pre_loop_and() and
                    self._pre_loop_list(code_gen, context_stack)):
                return False

        code_gen.add_instructions(inner_coder.program)
        return True

    def _push_light_names(self, code_gen, operand) -> bool:
        inner_code = CodeGen()
        inner_code.plus_equals(LoopVar.COUNTER)
        inner_code.push(LoopVar.CURRENT)
        if self._loop_type is _LoopType.LIST:
            if not self.rvalue(LoopVar.FIRST, code_gen):
                return self.token_error(
                    'Needed name of a group or location, got "{}"')
            code_gen.iter_members(operand, inner_code.program)
        else:
            code_gen.iter_lights(inner_code.program)
        return True

    def _push_set_names(self, code_gen) -> bool:
        """Generate code to push all group or location names onto the stack. """
        inner_code = CodeGen()
        inner_code.plus_equals(LoopVar.COUNTER)
        inner_code.push(LoopVar.CURRENT)
        if self._loop_type is _LoopType.GROUPS:
            operand = Operand.GROUP
        else:
            operand = Operand.LOCATION
        code_gen.iter_sets(operand, inner_code.program)
        return True

    def _pre_loop_with(self, code_gen, context_stack) -> bool:
        if not self._init_index_var(context_stack):
            return False
        if self.current_token.is_a(TokenTypes.IN):
            code_gen.add_instruction(OpCode.MOVEQ, 0, LoopVar.COUNTER)
            self._loop_type = _LoopType.LIST
            self.next_token()
            return self._pre_loop_list(code_gen, context_stack)
        if self.current_token.is_a(TokenTypes.FROM):
            self.next_token()
            if not self._index_var_range(code_gen):
                return False
        elif self.current_token.is_a(TokenTypes.CYCLE):
            self.next_token()
            if not self._cycle_var_range(code_gen):
                return False
        else:
            return self.token_error(
                'Needed "from" or "cycle", got "{}"')
        return True

    def _pre_loop_and(self) -> bool:
        self.next_token()
        if (self.current_token.token_type not in
            (TokenTypes.GROUP, TokenTypes.LOCATION)
                and not self.at_rvalue()):
            return self.token_error('Needed lights after "and", got "{}".')
        return True

    def _pre_loop_as(self, context_stack) -> bool:
        if not self.current_token.is_a(TokenTypes.AS):
            return self.token_error('Expected "as", got "{}"')
        self.next_token()
        if not self.current_token.is_a(TokenTypes.NAME):
            return self.token_error('Expected name for lights, got "{}"')
        self._light_var = str(self.current_token)
        context_stack.add_variable(self._light_var)
        return self.next_token()

    def _init_index_var(self, context_stack) -> bool:
        if not self.current_token.is_a(TokenTypes.NAME):
            return self.token_error('Not a variable name: "{}"')
        self._index_var = str(self.current_token)
        context_stack.add_variable(self._index_var)
        return self.next_token()

    def _index_var_range(self, code_gen) -> bool:
        if not self.rvalue(LoopVar.FIRST):
            return False
        if not self.current_token.is_a(TokenTypes.TO):
            return self.token_error('Needed "to", got "{}"')
        self.next_token()
        if not self.rvalue(LoopVar.LAST):
            return False
        code_gen.add_instruction(OpCode.MOVE, LoopVar.FIRST, self._index_var)
        if self._loop_type is _LoopType.WITH:
            if not self._calc_counter(code_gen):
                return False
        else:
            if not self._calc_incr(code_gen):
                return False
        return True

    @staticmethod
    def _calc_counter(code_gen) -> bool:
        # abs(last - first) + 1
        code_gen.subtract(LoopVar.LAST, LoopVar.FIRST)
        code_gen.pop(LoopVar.COUNTER)
        code_gen.test_op(Operator.LT, LoopVar.COUNTER, 0)
        marker = code_gen.if_true_start()
        code_gen.times_equals(LoopVar.COUNTER, -1)
        code_gen.add_instruction(OpCode.MOVEQ, -1, LoopVar.INCR)
        code_gen.if_else(marker)
        code_gen.add_instruction(OpCode.MOVEQ, 1, LoopVar.INCR)
        code_gen.if_end(marker)
        code_gen.plus_equals(LoopVar.COUNTER)
        return True

    @staticmethod
    def _calc_incr(code_gen) -> bool:
        # increment = (last - first) / (count - 1)
        # If count == 1, increment = 0.
        code_gen.test_op(Operator.NOTEQ, LoopVar.COUNTER, 1)
        marker = code_gen.if_true_start()
        code_gen.subtract(LoopVar.LAST, LoopVar.FIRST)
        code_gen.subtract(LoopVar.COUNTER, 1)
        code_gen.add_list(
            (OpCode.OP, Operator.DIV),
            (OpCode.POP, LoopVar.INCR)
        )
        code_gen.if_else(marker)
        code_gen.add_instruction(OpCode.MOVEQ, 0, LoopVar.INCR)
        code_gen.if_end(marker)
        return True

    def _cycle_var_range(self, code_gen):
        if self._loop_type == (_LoopType.WITH):
            return self.trigger_error(
                "repeat with cycle missing number of iterations")
        if not self.at_rvalue():
            code_gen.add_instruction(OpCode.MOVEQ, 0, LoopVar.FIRST)
        elif not self.rvalue(LoopVar.FIRST):
            return False

        # increment = 360 / counter, or
        # increment = 65536 / counter, based on unit_mode register
        code_gen.add_instruction(OpCode.MOVE, LoopVar.FIRST, self._index_var)
        code_gen.test_op(Operator.EQ, Register.UNIT_MODE, UnitMode.RAW)
        marker = code_gen.if_true_start()
        code_gen.push(65536)
        code_gen.if_else(marker)
        code_gen.push(360)
        code_gen.if_end(marker)
        code_gen.push(LoopVar.COUNTER)
        code_gen.add_instruction(OpCode.OP, Operator.DIV)
        code_gen.add_instruction(OpCode.POP, LoopVar.INCR)
        return True

    def _loop_test(self, code_gen) -> bool:
        '''
        Generate code to leave True or False in the result register,
        depending on whether the loop should continue.
        '''
        if self._loop_type is _LoopType.INFINITE:
            code_gen.add_instruction(OpCode.MOVEQ, True, Register.RESULT)
            return True
        if self._loop_type is _LoopType.WHILE:
            return self.rvalue()
        else:
            code_gen.test_op(Operator.GT, LoopVar.COUNTER, 0)
        return True

    def _loop_body(self, code_gen) -> bool:
        if self._loop_type.is_iter():
            code_gen.add_instruction(OpCode.POP, self._light_var)
        return self.parser.command_seq()

    def _loop_post(self, code_gen) -> bool:
        if self._loop_type in (_LoopType.INFINITE, _LoopType.WHILE):
            return True
        code_gen.minus_equals(LoopVar.COUNTER, 1)
        if self._index_var is not None:
            code_gen.plus_equals(self._index_var, LoopVar.INCR)
        return True

    @property
    def _current_operand_token(self):
        return {
            TokenTypes.GROUP: Operand.GROUP,
            TokenTypes.LOCATION: Operand.LOCATION
        }.get(self.current_token.token_type)



from bardolph.lib import injection, settings
from bardolph.controller import config_values, light_module
from bardolph.lib import job_control
from bardolph.controller.script_job import ScriptJob

class LsModule:
    _jobs = job_control.JobControl()

    @staticmethod
    def queue_script(script):
        return LsModule._jobs.add_job(ScriptJob.from_string(script))


def configure():
    injection.configure()
    settings.using(config_values.functional).apply_env().configure()
    light_module.configure()


def queue_script(script) -> job_control.Agent:
    return LsModule.queue_script(script) or job_control.failed_job()



from collections import deque

from .vm_codes import LoopVar

class StackFrame:
    def __init__(self, variables=None, return_addr=None):
        self.vars = variables or {}
        self.return_addr = return_addr or 0

    def clear(self):
        self.vars.clear()
        self.return_addr = 0

class LoopFrame(StackFrame):
    def __init__(self, variables):
        super().__init__(variables, None)
        self._loop_var = {}

    def get_loop_var(self, index):
        return self._loop_var.get(index, None)

    def set_loop_var(self, index, value):
        self._loop_var[index] = value

class CallStack:
    """
    As PARAM instructions are encountered, incoming parameters are saved in
    _current but are out of scope. That StackFrame instance gets pushed when a
    routine is called. The presence of that StackFram at the top of the stack
    brings the parameters into scope.

    Variables are immediately put into the StackFrame referenced by self.top.
    If the stack has only one StackFrame, then declared variables become
    globals. Otherwise, they are local variables that go out of scope when the
    routine returns.
    """

    def __init__(self):
        """
        Put an empty StackFrame into the stack as the root context. Create
        a new StackFrame to accumulate parameters during set-up for a routine
        invocation.
        """
        self._stack = deque()
        self._root_frame = StackFrame()
        self._stack.append(self._root_frame)
        self._current = StackFrame()
        self._constants = {}

    def reset(self) -> None:
        self._stack.clear()
        self._root_frame.clear()
        self._stack.append(self._root_frame)
        self._current.clear()
        self._constants.clear()

    @property
    def top(self) -> StackFrame:
        return self._stack[-1]

    def put_param(self, name, value) -> None:
        self._current.vars[name] = value

    def put_constant(self, name, value) -> None:
        if name not in self._constants:
            self._constants[name] = value

    def put_variable(self, index, value) -> None:
        if isinstance(index, LoopVar):
            assert isinstance(self.top, LoopFrame)
            return self.top.set_loop_var(index, value)
        dest = self._root_frame if index in self._root_frame.vars else self.top
        dest.vars[index] = value

    def is_param(self, name) -> bool:
        return name in self.top.vars

    def set_return(self, address) -> None:
        self._current.return_addr = address

    def get_return(self) -> int:
        return self.top.return_addr

    def get_variable(self, name):
        if isinstance(name, LoopVar):
            if isinstance(self.top, LoopFrame):
                return self.top.get_loop_var(name)
            return None
        for place in (self._constants, self.top.vars, self._root_frame.vars):
            if name in place:
                return place[name]
        return None

    def push_current(self) -> None:
        self._stack.append(self._current)
        self._current = StackFrame()

    def enter_loop(self) -> None:
        self._stack.append(LoopFrame(self.top.vars))

    def exit_loop(self) -> None:
        assert len(self._stack) > 1
        self._stack.pop()

    def unwind_loops(self) -> None:
        while isinstance(self.top, LoopFrame):
            self._stack.pop()

    def pop_current(self) -> None:
        assert len(self._stack) > 1
        self._current = self._stack.pop()
        self._current.clear()



from numbers import Number
import operator

from bardolph.controller.units import UnitMode

from .eval_stack import EvalStack
from .vm_codes import LoopVar, Operand, Operator, Register

class VmMath:
    _fn_table = { op_code: op_fn for op_code, op_fn in (
        (Operator.ADD, operator.add),
        (Operator.DIV, operator.truediv),
        (Operator.EQ, operator.__eq__),
        (Operator.GT, operator.gt),
        (Operator.GTE, operator.ge),
        (Operator.LT, operator.lt),
        (Operator.LTE, operator.le),
        (Operator.NOTEQ, operator.ne),
        (Operator.MOD, operator.mod),
        (Operator.MUL, operator.mul),
        (Operator.POW, operator.pow),
        (Operator.SUB, operator.sub)
    )}

    def __init__(self, call_stack, reg):
        self._call_stack = call_stack
        self._reg = reg
        self._eval_stack = EvalStack()

    def reset(self) -> None:
        self._eval_stack.clear()

    def push(self, srce) -> None:
        value = None
        if isinstance(srce, Register):
            value = self._reg.get_by_enum(srce)
        elif isinstance(srce, (Number, UnitMode)) or srce is Operand.NULL:
            value = srce
        elif isinstance(srce, (str, LoopVar)):
            value = self._call_stack.get_variable(srce)
        assert value is not None
        self._eval_stack.push(value)

    def pushq(self, srce) -> None:
        self._eval_stack.push(srce)

    def pop(self, dest) -> None:
        value = self._eval_stack.pop()
        if isinstance(dest, Register):
            self._reg.set_by_enum(dest, value)
        elif isinstance(dest, (str, LoopVar)):
            self._call_stack.put_variable(dest, value)

    def op(self, operator) -> None:
        if operator in (Operator.UADD, Operator.USUB, Operator.NOT):
            self.unary_op(operator)
        elif operator in (Operator.AND, Operator.OR):
            self.logical_op(operator)
        else:
            self.bin_op(operator)

    def unary_op(self, operator) -> None:
        if operator is Operator.USUB:
            self._eval_stack.replace_top(-self._eval_stack.top)
        elif operator is Operator.NOT:
            self._eval_stack.replace_top(not self._eval_stack.top)

    def bin_op(self, operator) -> None:
        op2 = self._eval_stack.pop()
        op1 = self._eval_stack.pop()
        self._eval_stack.push(VmMath._fn_table[operator](op1, op2))

    def logical_op(self, operator) -> None:
        op2 = bool(self._eval_stack.pop())
        op1 = bool(self._eval_stack.pop())
        result = op1 and op2 if operator == Operator.AND else op1 or op2
        self._eval_stack.push(result)



from bardolph.vm.vm_codes import Register

class SubParser:
    def __init__(self, parser):
        self.parser = parser

    @property
    def current_token(self):
        return self.parser.current_token

    @property
    def code_gen(self):
        return self.parser._code_gen

    @property
    def context(self):
        return self.parser._context

    @property
    def current_float(self):
        return self.parser._current_float()

    @property
    def current_str(self):
        return self.parser._current_str()

    @property
    def current_reg(self):
        return self.parser._current_reg()

    @property
    def current_literal(self):
        return self.parser._current_literal()

    def next_token(self):
        return self.parser.next_token()

    def rvalue(self, dest=Register.RESULT, code_gen=None):
        return self.parser._rvalue(dest, code_gen)

    def at_rvalue(self):
        return self.parser._at_rvalue()

    def trigger_error(self, msg):
        return self.parser.trigger_error(msg)

    def token_error(self, fmt):
        return self.parser.token_error(fmt)



#!/usr/bin/env python

import logging
import threading
import time

import lifxlan

from bardolph.lib.color import rounded_color
from bardolph.lib.injection import bind_instance, inject
from bardolph.lib.sorted_list import SortedList
from bardolph.lib.i_lib import Settings

from .i_controller import Lifx
from . import i_controller
from .light import Light

class LightSet(i_controller.LightSet):
    """
    The lights are stored in _light_dict, keyed on light name. Locations and
    groups are stored in _location_dict and _group_dict as lists of strings
    containing light names.
    """
    the_instance = None

    def __init__(self):
        self._light_dict = {}
        self._group_dict = {}
        self._location_dict = {}
        self._light_names = SortedList()
        self._num_successful_discovers = 0
        self._num_failed_discovers = 0

    @staticmethod
    def configure():
        LightSet.the_instance = LightSet()

    @staticmethod
    def get_instance():
        return LightSet.the_instance

    @inject(Lifx)
    def discover(self, lifx):
        logging.debug('start discover. so far, successes = {}, fails = {}'
                      .format(self._num_successful_discovers,
                              self._num_failed_discovers))
        try:
            for lifx_light in lifx.get_lights():
                light = Light(lifx_light)
                self._light_dict[light.name] = light
                self._light_names.add(light.name)
                LightSet._update_memberships(
                    light, light.group, self._group_dict)
                LightSet._update_memberships(
                    light, light.location, self._location_dict)
        except lifxlan.errors.WorkflowException as ex:
            self._num_failed_discovers += 1
            logging.warning("In discover():\n{}".format(ex))
            return False

        self._num_successful_discovers += 1
        return True

    def refresh(self):
        self.discover()
        self._garbage_collect()

    @staticmethod
    def _update_memberships(light, current_set_name, set_dict):
        LightSet._remove_memberships(light, set_dict)
        if current_set_name not in set_dict:
            set_dict[current_set_name] = SortedList(light.name)
        else:
            set_dict[current_set_name].add(light.name)

    @staticmethod
    def _remove_memberships(light, set_dict):
        # Remove the light from every set in set_dict that it belongs to.
        target_set_names = []
        for list_name in set_dict.keys():
            the_list = set_dict[list_name]
            the_list.remove(light.name)
            if len(the_list) == 0:
                target_set_names.append(list_name)
        for set_name in target_set_names:
            del set_dict[set_name]

    @inject(Settings)
    def _garbage_collect(self, settings):
        # Get rid of a light's proxy if it hasn't responded for a while.
        logging.debug("garbage collect, currently have {} lights"
                      .format(len(self._light_dict)))
        max_age = int(settings.get_value('light_gc_time', 20 * 60))
        target_lights = []
        for light in self._light_dict.values():
            if light.get_age() > max_age:
                LightSet._remove_memberships(light, self._group_dict)
                LightSet._remove_memberships(light, self._location_dict)
                target_lights.append(light.name)
        for light_name in target_lights:
            logging.debug("_garbage_collect() deleting {}".format(light_name))
            self._light_names.remove(light_name)
            self._light_dict[light_name] = None
            del self._light_dict[light_name]

    @property
    def light_names(self) -> SortedList:
        """ list of strings """
        return self._light_names

    @property
    def group_names(self):
        """ list of strings """
        return SortedList(self._group_dict.keys())

    @property
    def location_names(self):
        """ list of strings """
        return SortedList(self._location_dict.keys())

    @property
    def count(self):
        return len(self._light_list)

    @property
    def successful_discovers(self):
        return self._num_successful_discovers

    @property
    def failed_discovers(self):
        return self._num_failed_discovers

    def get_light(self, name):
        """ returns an instance of i_lib.Light, or None if it's not there """
        return self._light_dict.get(name)

    def get_group(self, name):
        """ list of light names """
        return self._group_dict.get(name)

    def get_location(self, name):
        """ list of light names. """
        return self._location_dict.get(name)

    @inject(Lifx)
    def set_color(self, color, duration, lifx):
        lifx.set_color_all_lights(rounded_color(color), duration)
        return True

    @inject(Lifx)
    def set_power(self, power_level, duration, lifx):
        lifx.set_power_all_lights(round(power_level), duration)
        return True


def start_light_refresh():
    logging.debug("Starting refresh thread.")
    threading.Thread(
        target=light_refresh, name='rediscover', daemon=True).start()


@inject(Settings)
def light_refresh(settings):
    success_sleep_time = float(
        settings.get_value('refresh_sleep_time', 600))
    failure_sleep_time = float(
        settings.get_value('failure_sleep_time', success_sleep_time))
    complete_success = False

    while True:
        time.sleep(
            success_sleep_time if complete_success else failure_sleep_time)
        lights = LightSet.get_instance()
        try:
            complete_success = lights.refresh()
        except lifxlan.errors.WorkflowException as ex:
            logging.warning("Error during discovery {}".format(ex))


@inject(Settings)
def configure(settings):
    LightSet.configure()
    lights = LightSet.get_instance()
    bind_instance(lights).to(i_controller.LightSet)
    lights.discover()

    single = bool(settings.get_value('single_light_discover', False))
    if not single:
        start_light_refresh()



class Trace:
    trace_enabled = False
    callback = print

def trace_call(fn):
    def wrapper(*args, **kwargs):
        if not Trace.trace_enabled:
            return fn(*args, **kwargs)

        Trace.callback("Enter: {}".format(fn.__name__))
        result = fn(*args, **kwargs)
        Trace.callback("Exit: {}".format(fn.__name__))
        return result
    return wrapper

def trace_call_enable(enable):
    Trace.trace_enabled = enable

def trace_call_callback(callback):
    Trace.callback = callback



from enum import Enum, auto


class Assoc(Enum):
    LEFT = auto()
    RIGHT = auto()


class TokenTypes(Enum):
    ALL = auto()
    AND = auto()
    AS = auto()
    ASSIGN = auto()
    AT = auto()
    BEGIN = auto()
    BREAK = auto()
    BREAKPOINT = auto()
    COMPARE = auto()
    CYCLE = auto()
    DEFINE = auto()
    ELSE = auto()
    END = auto()
    EOF = auto()
    ERROR = auto()
    FROM = auto()
    GET = auto()
    GROUP = auto()
    IF = auto()
    IN = auto()
    LITERAL_STRING = auto()
    LOCATION = auto()
    LOGICAL = auto()
    MARK = auto()
    NAME = auto()
    NOT = auto()
    NULL = auto()
    NUMBER = auto()
    OFF = auto()
    ON = auto()
    OR = auto()
    PRINT = auto()
    PRINTF = auto()
    PRINTLN = auto()
    PAUSE = auto()
    RAW = auto()
    REGISTER = auto()
    REPEAT = auto()
    RETURN = auto()
    RGB = auto()
    SET = auto()
    SYNTAX_ERROR = auto()
    TIME_PATTERN = auto()
    TO = auto()
    UNITS = auto()
    UNKNOWN = auto()
    WHILE = auto()
    WITH = auto()
    WAIT = auto()
    ZONE = auto()

    def has_string(self):
        return self in (TokenTypes.ERROR, TokenTypes.LITERAL_STRING,
            TokenTypes.MARK, TokenTypes.NAME, TokenTypes.NUMBER,
            TokenTypes.REGISTER, TokenTypes.TIME_PATTERN)

    def is_executable(self):
        return self in (
            TokenTypes.ASSIGN, TokenTypes.BREAKPOINT,
            TokenTypes.GET, TokenTypes.IF, TokenTypes.OFF, TokenTypes.ON,
            TokenTypes.PRINT, TokenTypes.PRINTF, TokenTypes.PRINTLN,
            TokenTypes.PAUSE, TokenTypes.REGISTER, TokenTypes.REPEAT,
            TokenTypes.SET, TokenTypes.UNITS, TokenTypes.WHILE, TokenTypes.WAIT)

class Token:
    def __init__(self,
            token_type, content='', line_number=0, file_name=''):
        self._token_type = token_type
        self._content = content
        self._line_number = line_number
        self._file_name = file_name

    def __eq__(self, other):
        if isinstance(other, str):
            if self._token_type.has_string():
                return self._content == other
            else:
                return False
        if isinstance(other, TokenTypes):
            return (not self._token_type.has_string()
                and self._token_type is other)
        if isinstance(other, Token):
            if self._token_type is not other._token_type:
                return False
            if not self._token_type.has_string():
                return True
        return self._content == other._content

    def __repr__(self):
        fmt = "Token({}"
        if len(self._content) > 0:
            fmt += ", '{}'"
        if self._line_number > 0:
            fmt += ", {}"
        if len(self._file_name) > 0:
            fmt += ", '{}'"
        fmt += ')'
        return fmt.format(
            self._token_type, self._content, self._line_number, self._file_name)

    def __str__(self):
        if self._token_type.has_string():
            return self._content
        return self._token_type.name.lower()

    def is_a(self, token_type) -> bool:
        return self._token_type is token_type

    @property
    def token_type(self):
        return self._token_type

    @property
    def content(self):
        return self._content

    @property
    def file_name(self):
        return self.file_name

    @property
    def is_binop(self):
        return (self.is_a(TokenTypes.COMPARE)
                or self.content in '+-*/^'
                or self.content in ('and', 'or'))

    @property
    def line_number(self):
        return self._line_number

    @property
    def prec(self):
        return {
            'not': 1,
            'or': 2,
            'and': 3,
            '==': 4,
            '<=': 4,
            '>=': 4,
            '!=': 4,
            '<': 4,
            '>': 4,
            '+': 5,
            '-': 5,
            '*': 6,
            '/': 6,
            '^': 7
        }.get(self.content, -1)

    @property
    def assoc(self):
        if self.content in ('not', '^'):
            return Assoc.RIGHT
        return Assoc.LEFT



import logging

def add_n_argument(parser):
    parser.add_argument(
        '-n', '--num-lights', help='number of lights to discover, >= 0')
    
def add_o_argument(parser):
    parser.add_argument(
        '-o', '--output-file',
        help="name of the output file (default is stdout)", nargs='+')
    
def get_output_file(args):
    return None if args.output_file is None else args.output_file[0]

def get_overrides(args):
    if args.num_lights is None:
        return None
        
    try:
        num_lights = int(args.num_lights)
        if num_lights < 0:
            raise ValueError
    except ValueError:
        logging.error("Invalid value for -n or --num-lights.")
        logging.error("Must be a integer greater than or equal to zero.")
        raise ValueError

    return {'default_num_lights': num_lights}



from collections import deque

from bardolph.lib.symbol import Symbol, SymbolType
from bardolph.lib.symbol_table import Symbol, SymbolTable


class _LoopContext:
    def __init__(self):
        self.break_list = []


class Context:
    def __init__(self):
        self._globals = SymbolTable()
        self._locals = SymbolTable()
        self._loop_stack = deque()
        self._loop_depth = 0
        self._in_routine = False
        self._return_list = []

    def __contains__(self, name) -> bool:
        return name in self._locals or name in self._globals

    def clear(self) -> None:
        self._in_routine = False
        self._globals.clear()
        self._locals.clear()
        self._loop_stack.clear()
        self._return_list.clear()

    def enter_routine(self) -> None:
        self._in_routine = True

    def in_routine(self) -> bool:
        return self._in_routine

    def exit_routine(self) -> None:
        self._in_routine = False
        self._locals.clear()
        self._return_list.clear()

    def enter_loop(self) -> None:
        self._loop_stack.append(_LoopContext())

    def in_loop(self) -> bool:
        return len(self._loop_stack) > 0

    def exit_loop(self) -> None:
        self._loop_stack.pop()

    def add_break(self, inst) -> None:
        self._top_break_list().append(inst)

    def _top_break_list(self):
        return self._loop_stack[-1].break_list

    def fix_break_addrs(self, code_gen) -> None:
        offset = code_gen.current_offset
        for inst in self._top_break_list():
            inst.param1 = offset - inst.param1

    def add_return(self, inst) -> None:
        self._return_list.append(inst)

    def fix_return_addrs(self, code_gen) -> None:
        offset = code_gen.current_offset
        for inst in self._return_list:
            inst.param1 = offset - inst.param1

    def add_routine(self, routine) -> None:
        self._globals.add_symbol(routine.name, SymbolType.ROUTINE, routine)

    def add_variable(self, name, value=None) -> None:
        dest = self._locals if self._in_routine else self._globals
        dest.add_symbol(name, SymbolType.VAR, value)

    def add_global(self, name, symbol_type, value) -> None:
        self._globals.add_symbol(name, symbol_type, value)

    def get_data(self, name):
        return self.get_symbol_typed(name, (SymbolType.MACRO, SymbolType.VAR))

    def get_symbol(self, name) -> Symbol:
        """
        Get a parameter from the top of the stack. If it's not there, check
        the globals.
        """
        symbol = self._locals.get_symbol(name)
        if symbol.undefined:
            symbol = self._globals.get_symbol(name)
        return symbol

    def get_symbol_typed(self, name, symbol_types):
        symbol = self.get_symbol(name)
        if symbol.undefined or symbol.symbol_type in symbol_types:
            return symbol
        return Symbol()

    def has_symbol(self, name) -> bool:
        return not self.get_symbol(name).undefined

    def has_symbol_typed(self, name, * symbol_types) -> bool:
        symbol = self.get_symbol(name)
        return not symbol.undefined and symbol.symbol_type in symbol_types

    def get_routine(self, name):
        return self._global_of_type(name, SymbolType.ROUTINE)

    def has_routine(self, name):
        return not self._global_of_type(name, SymbolType.ROUTINE).undefined

    def get_macro(self, name) -> Symbol:
        return self._global_of_type(name, SymbolType.MACRO)

    def _global_of_type(self, name, symbol_type) -> Symbol:
        symbol = self._globals.get_symbol(name)
        if symbol.undefined or symbol.symbol_type == symbol_type:
            return symbol
        return Symbol()



#!/usr/bin/env python

import argparse
import os
import stat

from . import arg_helper
from ..lib import injection
from ..lib import settings
from ..parser.parse import Parser

from . import config_values

def program_code(instructions):
    output = ''
    dot = os.path.dirname(os.path.realpath(__file__))
    with open(os.path.join(dot, 'lsc_template.py')) as srce:
        for line in srce:
            if line.find('#instructions') > -1:
                output += instructions
            else:
                output += line
    return output

def instruction_text(file_name):
    parser = Parser()
    program = parser.load(file_name)
    if program is None:
        print("Error compiling {}".format(file_name))
        print(parser.get_errors())
        return None
    text = '    '
    text += ',\n    '.join(map(lambda inst: inst.as_list_text(), program))
    return text

def output_python(output_text, output_name=None):
    if output_name is None:
        print(output_text)
    else:
        out_file = open(output_name, 'w')
        out_file.write(output_text)
        out_file.close()
        os.chmod(output_name, stat.S_IRWXU | stat.S_IRGRP | stat.S_IROTH)

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('file', help='name of the script file')
    arg_helper.add_o_argument(parser)
    args = parser.parse_args()

    injection.configure()
    settings.using(config_values.functional).apply_env().configure()

    input_file = args.file
    program = instruction_text(input_file)
    if program is not None:
        output_python(program_code(program), arg_helper.get_output_file(args))

if __name__ == '__main__':
    main()



import os

from .vm_codes import OpCode


class Instruction:
    def __init__(self, op_code, param0=None, param1=None):
        self.op_code = op_code
        self.param0 = param0
        self.param1 = param1

    def __repr__(self):
        if self.op_code is OpCode.TIME_PATTERN:
            return 'Instruction({}, {}, {})'.format(
                OpCode.TIME_PATTERN, self.param0, repr(self.param1))
        if self.param1 is None:
            if self.param0 is None:
                return 'Instruction({})'.format(self.op_code)
            return 'Instruction({}, {})'.format(
                self.op_code,
                Instruction.quote_if_string(self.param0))
        return 'Instruction({}, {}, {})'.format(
            self.op_code,
            Instruction.quote_if_string(self.param0),
            Instruction.quote_if_string(self.param1))

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            raise TypeError
        return (self.op_code == other.op_code
                and self.param0 == other.param0
                and self.param1 == other.param1)

    def nop(self):
        self.op_code = OpCode.NOP

    def as_list_text(self) -> str:
        if self.param0 is None and self.param1 is None:
            return str(self.op_code)
        if self.param1 is None:
            return '{}, {}'.format(
                self.op_code,
                Instruction.quote_if_string(self.param0))
        if self.param1 is os.linesep:
            param1 = repr(os.linesep)
        else:
            param1 = Instruction.quote_if_string(self.param1)
        return '{}, {}, {}'.format(
            self.op_code, Instruction.quote_if_string(self.param0), param1)

    @staticmethod
    def quote_if_string(obj):
        return ('"{}"' if isinstance(obj, str) else '{}').format(obj)

    @staticmethod
    def do_listing(program):
        result = ''
        inst_num = 0
        for inst in program:
            result += '{:5d} {}\n'.format(inst_num, inst)
            inst_num += 1
        return result



#!/usr/bin/env python

import argparse
import logging

from bardolph.controller.routine import Routine
from bardolph.controller.units import UnitMode
from bardolph.lib.symbol_table import SymbolType
from bardolph.lib.time_pattern import TimePattern
from bardolph.parser.expr_parser import ExpressionParser
from bardolph.vm.vm_codes import (JumpCondition, OpCode, Operand, Operator,
                                  Register, SetOp)

from .code_gen import CodeGen
from .context import Context
from .io_parser import IoParser
from .lex import Lex
from .loop_parser import LoopParser
from .token import Token, TokenTypes


class Parser:
    def __init__(self):
        self._lexer = None
        self._error_output = ''
        self._context = Context()
        self._current_token = Token(TokenTypes.UNKNOWN)
        self._op_code = OpCode.NOP
        self._code_gen = CodeGen()
        self._tokens = iter([])
        self._command_map = {
            TokenTypes.ASSIGN: self._assignment,
            TokenTypes.BREAK: self._break,
            TokenTypes.BREAKPOINT: self._breakpoint,
            TokenTypes.DEFINE: self._definition,
            TokenTypes.GET: self._get,
            TokenTypes.IF: self._if,
            TokenTypes.MARK: self._mark,
            TokenTypes.NAME: self._call_routine,
            TokenTypes.NULL: self._syntax_error,
            TokenTypes.OFF: self._power_off,
            TokenTypes.ON: self._power_on,
            TokenTypes.PAUSE: self._pause,
            TokenTypes.PRINT: self._print,
            TokenTypes.PRINTF: self._printf,
            TokenTypes.PRINTLN: self._println,
            TokenTypes.RETURN: self._return,
            TokenTypes.REGISTER: self._set_reg,
            TokenTypes.REPEAT: self._repeat,
            TokenTypes.SET: self._set,
            TokenTypes.UNITS: self._set_units,
            TokenTypes.WAIT: self._wait,
            None: self._syntax_error
        }
        self._token_trace = False

    def parse(self, input_string):
        self._context.clear()
        self._code_gen.clear()
        self._error_output = ''
        self._tokens = Lex(input_string).tokens()
        self.next_token()
        success = self._script()
        return self._code_gen.program if success else None

    def load(self, file_name):
        logging.debug('File name: {}'.format(file_name))
        try:
            srce = open(file_name, 'r')
            input_string = srce.read()
            srce.close()
            return self.parse(input_string)
        except FileNotFoundError:
            logging.error('Error: file {} not found.'.format(file_name))
        except OSError:
            logging.error('Error accessing file {}'.format(file_name))

    def get_errors(self) -> str:
        return self._error_output

    @property
    def current_token(self):
        return self._current_token

    def _script(self) -> bool:
        return self._body() and self._eof()

    def _body(self) -> bool:
        while not self._current_token.is_a(TokenTypes.EOF):
            if not self._command():
                return False
        return True

    def _eof(self) -> bool:
        if not self._current_token.is_a(TokenTypes.EOF):
            return self.trigger_error("Didn't get to end of file.")
        return True

    def _command(self):
        return self._command_map.get(
            self._current_token.token_type, self._syntax_error)()

    def _set_reg(self):
        reg = Register.from_string(str(self._current_token))
        if reg is None:
            return self.token_error('Expected register, got "{}"')
        if reg is Register.TIME:
            return self._time()

        self.next_token()
        if self._current_token.is_a(TokenTypes.LITERAL_STRING):
            return self._string_to_reg(reg)
        return self._rvalue(reg)

    def _string_to_reg(self, reg) -> bool:
        if reg != Register.NAME:
            return self.trigger_error('Quoted value not allowed here.')
        self._add_instruction(OpCode.MOVEQ, self._current_token.content, reg)
        return self.next_token()

    def _set(self):
        return self._action(OpCode.COLOR)

    def _power_on(self):
        self._add_instruction(OpCode.MOVEQ, True, Register.POWER)
        return self._action(OpCode.POWER)

    def _power_off(self) -> bool:
        self._add_instruction(OpCode.MOVEQ, False, Register.POWER)
        return self._action(OpCode.POWER)

    def _action(self, op_code) -> bool:
        self._op_code = op_code
        self._add_instruction(OpCode.WAIT)
        self.next_token()
        if self._current_token.is_a(TokenTypes.ALL):
            return self._all_operand()
        return self._operand_list()

    def _all_operand(self) -> bool:
        self._add_instruction(OpCode.MOVEQ, None, Register.NAME)
        self._add_instruction(OpCode.MOVEQ, Operand.ALL, Register.OPERAND)
        self._add_instruction(self._op_code)
        return self.next_token()

    def _operand_list(self) -> bool:
        """
        For every operand in the list, issue the instruction in
        self._op_code.
        """
        if not self._operand():
            return False
        self._add_instruction(self._op_code)

        while self._current_token.is_a(TokenTypes.AND):
            self.next_token()
            if not self._operand():
                return False
            self._add_instruction(self._op_code)
        return True

    def _operand(self) -> bool:
        """
        Process a group, location, or light with an optional set of
        zones. Do this by populating the NAME and OPERAND registers.
        """
        if self._current_token.is_a(TokenTypes.GROUP):
            operand = Operand.GROUP
            self.next_token()
        elif self._current_token.is_a(TokenTypes.LOCATION):
            operand = Operand.LOCATION
            self.next_token()
        else:
            operand = Operand.LIGHT

        const_str = self._current_str()
        if len(const_str) > 0:
            self._add_instruction(OpCode.MOVEQ, const_str, Register.NAME)
            self.next_token()
        elif self._current_token.is_a(TokenTypes.NAME):
            if not self._var_operand():
                return False
        else:
            return self.token_error('Needed a light, got "{}".')

        if self._current_token.is_a(TokenTypes.ZONE):
            if not self._zone_range():
                return False
            operand = Operand.MZ_LIGHT

        self._add_instruction(OpCode.MOVEQ, operand, Register.OPERAND)
        return True

    def _var_operand(self) -> bool:
        name = str(self._current_token)
        if not self._context.has_symbol_typed(
                name, SymbolType.MACRO, SymbolType.VAR):
            return self.token_error('Neither variable nor macro: "{}"')
        self._add_instruction(OpCode.MOVE, name, Register.NAME)
        return self.next_token()

    def _zone_range(self) -> bool:
        if self._op_code is not OpCode.COLOR:
            return self.trigger_error('Zones not supported for {}'.format(
                self._op_code.name.lower()))
        self.next_token()
        return self._set_zones()

    def _set_zones(self, only_one=False):
        """
        Parses out one or two zone numbers. Generates instructions
        that populate the FIRST_ZONE and LAST_ZONE registers. If only one
        parameter is present, put None into LAST_ZONE.
        """
        if not self._at_rvalue():
            return self.token_error('Expected zone, got "{}"')
        if not self._rvalue(Register.FIRST_ZONE):
            return False
        if not only_one and self._at_rvalue():
            return self._rvalue(Register.LAST_ZONE)

        self._add_instruction(OpCode.MOVEQ, None, Register.LAST_ZONE)
        return True

    def _set_units(self) -> bool:
        self.next_token()
        if self._current_token.token_type not in (
                TokenTypes.RAW, TokenTypes.RGB, TokenTypes.LOGICAL):
            return self.token_error('Invalid parameter "{}" for units.')
        mode = UnitMode[self._current_token.token_type.name]
        self._add_instruction(OpCode.MOVEQ, mode, Register.UNIT_MODE)
        return self.next_token()

    def _wait(self) -> bool:
        self._add_instruction(OpCode.WAIT)
        return self.next_token()

    def _get(self) -> bool:
        self.next_token()
        if not self._at_rvalue():
            return self.token_error('Needed light for get, got "{}".')
        if not self._rvalue(Register.NAME):
            return False

        if self._current_token.is_a(TokenTypes.ZONE):
            self.next_token()
            if not self._set_zones(True):
                return False
            operand = Operand.MZ_LIGHT
        else:
            operand = Operand.LIGHT

        self._add_instruction(OpCode.MOVEQ, operand, Register.OPERAND)
        self._add_instruction(OpCode.GET_COLOR)
        return True

    def _pause(self):
        self._add_instruction(OpCode.PAUSE)
        self.next_token()
        return True

    def _print(self) -> bool:
        return IoParser(self).print()

    def _printf(self) -> bool:
        return IoParser(self).printf()

    def _println(self) -> bool:
        return IoParser(self).println()

    def _time(self):
        self.next_token()
        if self._current_token.is_a(TokenTypes.AT):
            self.next_token()
            return self._process_time_patterns()
        return self._rvalue(Register.TIME)

    def _process_time_patterns(self):
        time_pattern = self._current_time_pattern()
        if time_pattern is None:
            return self._time_spec_error()
        self._add_instruction(
            OpCode.TIME_PATTERN, SetOp.INIT, time_pattern)
        self.next_token()

        while self._current_token.is_a(TokenTypes.OR):
            self.next_token()
            time_pattern = self._current_time_pattern()
            if time_pattern is None:
                return self._time_spec_error()
            self._add_instruction(
                OpCode.TIME_PATTERN, SetOp.UNION, time_pattern)
            self.next_token()

        return True

    def _assignment(self) -> bool:
        self.next_token()
        if not self._current_token.is_a(TokenTypes.NAME):
            return self.token_error('Expected name for assignment, got "{}"')
        dest_name = str(self._current_token)
        self.next_token()
        if not self._rvalue(dest_name):
            return False
        self._context.add_variable(dest_name)
        return True

    def _rvalue(self, dest=Register.RESULT, code_gen=None) -> bool:
        """
        Consume the current token as an rvalue, generating the code to evaluate
        it and to move the result into dest.
        """
        code_gen = code_gen or self._code_gen
        if self._current_token.content == '{':
            return self.next_token() and self._rvalue_curly(dest, code_gen)
        move_inst = OpCode.MOVE
        value = self._current_constant()
        if value is not None:
            move_inst = OpCode.MOVEQ
        elif self._current_token.is_a(TokenTypes.NAME):
            name = str(self._current_token)
            if self._context.has_symbol_typed(name, SymbolType.VAR):
                value = name
            else:
                if self._context.has_symbol(name):
                    return self.token_error('Not a value: "{}"')
                return self.token_error('Unknown: "{}"')
        elif self._current_token.is_a(TokenTypes.REGISTER):
            value = self._current_reg()
        elif self._current_token.content == 'not':
            return self._rvalue_not(dest, code_gen)
        else:
            return self.token_error('Cannot use {} as a value.')

        if dest is OpCode.PUSH:
            code_gen.push(value)
        else:
            code_gen.add_instruction(move_inst, value, dest)

        return self.next_token()

    def _rvalue_curly(self, dest, code_gen):
        if not self._rvalue_expr(dest, code_gen):
            return False
        if self.current_token != '}':
            return self.token_error("Expected closing curly brace, got {}.")
        return self.next_token()

    def _rvalue_not(self, dest, code_gen):
        self.next_token()
        if not ExpressionParser(self).expression():
            return False
        code_gen.add_instruction(OpCode.OP, Operator.NOT)
        return True

    def _rvalue_expr(self, dest, code_gen):
        if not ExpressionParser(self).expression():
            return False
        code_gen.pop(dest)
        return True

    def _at_rvalue(self) -> bool:
        if str(self.current_token) == '{':
            return True
        if self._current_token.token_type in (
                TokenTypes.LITERAL_STRING, TokenTypes.NUMBER):
            return True
        if self._current_token.is_a(TokenTypes.NAME):
            return not self._context.has_routine(str(self.current_token))
        return False

    def _definition(self) -> bool:
        self.next_token()
        if not self._current_token.is_a(TokenTypes.NAME):
            return self.token_error('Expected name for definition, got: {}')

        name = str(self._current_token)
        self.next_token()
        if self._detect_routine_start():
            if not self._context.get_routine(name).undefined:
                return self.token_error('Already defined: "{}"')
            return self._routine_definition(name)
        return self._macro_definition(name)

    def _detect_routine_start(self) -> bool:
        """
        If a definition is followed by "with", "begin", a keyword corresponding
        to a command, or the name of an existing routine, it's defining a new
        routine and not a variable.
        """
        if self._context.has_routine(str(self._current_token)):
            return True
        if self._current_token.token_type.is_executable():
            return True
        return self._current_token.token_type in (
            TokenTypes.BEGIN, TokenTypes.WITH)

    def _macro_definition(self, name):
        """
        Process a "define" where an alias for a value is being created. This
        symbol exists at compile time. This means a define cannot refer to a
        parameter in a routine. The symbol has global scope, even if it is
        defined inside a routine.
        """
        value = self._current_literal()
        if value is None:
            inner_macro = self._context.get_macro(str(self._current_token))
            if inner_macro is None:
                return self.token_error('Macro needs constant, got "{}"')
            value = inner_macro.value
        self._context.add_global(name, SymbolType.MACRO, value)
        self._add_instruction(OpCode.CONSTANT, name, value)
        return self.next_token()

    def _routine_definition(self, name):
        if self._context.in_routine():
            return self.trigger_error('Nested definition not allowed.')

        self._context.enter_routine()
        self._add_instruction(OpCode.ROUTINE, name)

        routine = Routine(name)
        self._context.add_routine(routine)
        if self._current_token.is_a(TokenTypes.WITH):
            self.next_token()
            if not self._param_decl(routine):
                return False
        result = self.command_seq()
        self._context.fix_return_addrs(self._code_gen)
        self._add_instruction(OpCode.END, name)
        self._context.exit_routine()
        return result

    def _param_decl(self, routine):
        """
        The parameter declarations for the routine are not included in the
        generated code. Declarations are used only at compile time.
        """
        name = str(self._current_token)
        routine.add_param(name)
        self._context.add_variable(name)
        self.next_token()
        while (self._current_token.is_a(TokenTypes.NAME) and not
                self._context.has_routine(str(self._current_token))):
            name = str(self._current_token)
            if routine.has_param(name):
                self.token_error('Duplicate parameter name: "{}"')
                return None
            routine.add_param(name)
            self._context.add_variable(name)
            self.next_token()
        return True

    def command_seq(self) -> bool:
        if not self._current_token.is_a(TokenTypes.BEGIN):
            return self._command()
        return self.compound_command()

    def compound_command(self) -> bool:
        self.next_token()
        while not self._current_token.is_a(TokenTypes.END):
            if self._current_token.is_a(TokenTypes.EOF):
                return self.trigger_error('End of file before "end".')
            if not self._command():
                return False
        return self.next_token()

    def _call_routine(self) -> bool:
        if str(self._current_token) == '[':
            self.next_token()
            bracketed = True
        else:
            bracketed = False
        routine = self._context.get_routine(str(self._current_token))
        if routine.undefined:
            return self.token_error('Unknown name: "{}"')

        self.next_token()
        for param_name in routine.value.params:
            if not self._rvalue():
                return False
            self._add_instruction(OpCode.PARAM, param_name, Register.RESULT)

        self._add_instruction(OpCode.JSR, routine.name)
        if bracketed:
            if str(self.current_token) != ']':
                return self.trigger_error(
                    'No closing bracket for function call.')
            self.next_token()
        return True

    def _call_fn(self, dest, code_gen) -> bool:
        if not self._call_routine():
            return False
        if dest is not Register.RESULT:
            code_gen.add_instruction(OpCode.MOVE, Register.RESULT, dest)
        return True

    def _return(self) -> bool:
        self.next_token()
        if not self._rvalue():
            return False
        inst = self._code_gen.add_instruction(
            OpCode.JUMP, JumpCondition.ALWAYS, self._code_gen.current_offset)
        self._context.add_return(inst)
        return True

    def _mark(self):
        # In this context, the result of a freestanding expression is thrown away.
        if str(self.current_token) != '{':
            return self.token_error("Unexpected character {}")
        return self.next_token() and self._expr()

    def _if(self) -> bool:
        self.next_token()
        if not self._rvalue():
            return False
        marker = self._code_gen.if_true_start()
        if not self.command_seq():
            return False
        if self.current_token.is_a(TokenTypes.ELSE):
            self._code_gen.if_else(marker)
            self.next_token()
            if not self.command_seq():
                return False
        self._code_gen.if_end(marker)
        return True

    def _repeat(self) -> bool:
        result = LoopParser(self).repeat(self._code_gen, self._context)
        return result

    def _break(self) -> bool:
        if not self._context.in_loop():
            return self.trigger_error('Encountered "break" not inside loop.')
        inst = self._code_gen.add_instruction(
            OpCode.JUMP, JumpCondition.ALWAYS, self._code_gen.current_offset)
        self._context.add_break(inst)
        return self.next_token()

    def _add_instruction(self, op_code, param0=None, param1=None):
        return self._code_gen.add_instruction(op_code, param0, param1)

    def _add_message(self, message):
        self._error_output += '{}\n'.format(message)

    def trigger_error(self, message):
        full_message = 'Line {}: {}'.format(
            self._current_token.line_number, message)
        self._add_message(full_message)
        return False

    def _current_literal(self):
        """
        Interpret the current token as a literal and return its value. If the
        current token doesn't contain a literal, return None.
        """
        value = None
        text = str(self._current_token)
        if self._current_token.is_a(TokenTypes.NUMBER):
            value = int(text) if Lex.is_int(text) else float(text)
        elif self._current_token.is_a(TokenTypes.LITERAL_STRING):
            value = str(self._current_token)
        elif self._current_token.is_a(TokenTypes.TIME_PATTERN):
            value = TimePattern.from_string(str(self._current_token))
            if value is None:
                self._time_spec_error()
        return value

    def _current_constant(self):
        """
        Interpret the current token as either a literal or macro and return
        its value, which is known at compile time. If the token is an undefined
        name, return None.
        """
        value = self._current_literal()
        if value is not None:
            return value
        if not self._current_token.is_a(TokenTypes.NAME):
            return None
        macro = self._context.get_macro(str(self._current_token))
        return None if macro.undefined else macro.value

    def _current_int(self):
        value = self._current_constant()
        if isinstance(value, int):
            return value
        return round(value) if isinstance(value, float) else None

    def _current_float(self):
        value = self._current_constant()
        if isinstance(value, float):
            return value
        return float(value) if isinstance(value, int) else None

    def _current_str(self) -> str:
        value = self._current_constant()
        return value if isinstance(value, str) else ''

    def _current_time_pattern(self) -> TimePattern:
        """
        Returns the current token as a time pattern. Only literals or macros.
        """
        if self._current_token.is_a(TokenTypes.TIME_PATTERN):
            return TimePattern.from_string(str(self._current_token))
        if self._current_token.is_a(TokenTypes.NAME):
            return self._context.get_macro(str(self._current_token)).value
        return TimePattern(None, None)

    def _current_reg(self):
        if not self._current_token.is_a(TokenTypes.REGISTER):
            return None
        return Register.from_string(str(self._current_token))

    def next_token(self):
        if self._current_token != TokenTypes.EOF:
            try:
                self._current_token = next(self._tokens)
            except StopIteration:
                self._current_token = Token(TokenTypes.EOF)
                return self.trigger_error('Unexpected end of source.')
        if self._token_trace:
            logging.info(
                'Next token: "{}" ({})'.format(
                    self._current_token, self._current_token.token_type))
        return True

    def _breakpoint(self):
        self._code_gen.add_instruction(OpCode.BREAKPOINT)

    def token_error(self, message_format):
        return self.trigger_error(
            message_format.format(str(self._current_token)))

    def _unimplementd(self):
        return self.token_error('Unimplemented at token "{}"')

    def _syntax_error(self):
        return self.token_error('Unexpected input "{}"')

    def _time_spec_error(self):
        return self.token_error('Invalid time specification: "{}"')


def main():
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument('file', help='name of the script file')
    args = arg_parser.parse_args()

    logging.basicConfig(
        level=logging.DEBUG,
        format='%(filename)s(%(lineno)d) %(funcName)s(): %(message)s')
    parser = Parser()
    output_code = parser.load(args.file)
    if output_code:
        inst_num = 0
        for inst in output_code:
            print('{:5d} {}'.format(inst_num, inst))
            inst_num += 1
    else:
        print("Error parsing: {}".format(parser.get_errors()))


if __name__ == '__main__':
    main()



from enum import Enum, auto

class SymbolType(Enum):
    EXTERN = auto()
    MACRO = auto()
    NO_TYPE = auto()
    PARAM = auto()
    ROUTINE = auto()
    UNDEFINED = auto()
    UNKNOWN = auto()
    VAR = auto()

class Symbol:
    def __init__(self, name='', symbol_type=SymbolType.UNDEFINED, value=None):
        self._name = name
        self._symbol_type = symbol_type
        self._value = value

    def __repr__(self):
        return 'Symbol("{}", {}, {})'.format(
            self.name, self.symbol_type, self.value)

    @property
    def undefined(self) -> bool:
        return self._symbol_type == SymbolType.UNDEFINED

    @property
    def name(self):
        return self._name

    @property
    def symbol_type(self):
        return self._symbol_type

    @property
    def value(self):
        return self._value



import collections
import logging
import threading


class Job:
    def execute(self): pass
    def request_stop(self): pass


class Agent:
    """
    The name serves as the unique identifier. When the job finishes, the
    callback is invoked with self (this Agent) as the only parameter.
    """
    def __init__(self, job, callback, name=None):
        self._job = job
        self._callback = callback
        self._thread = None
        self._name = name or 'job {}'.format(id(self))

    @property
    def name(self):
        return self._name

    @property
    def job(self):
        return self._job

    def is_running(self):
        return self._thread is not None and self._thread.is_alive()

    def execute(self):
        self._thread = threading.Thread(target=self._execute_and_call)
        self._thread.start()
        return self

    def request_stop(self):
        self._job.request_stop()

    def _execute_and_call(self):
        try:
            self._job.execute()
        finally:
            self._callback(self)


def failed_job() -> Agent:
    class FailedJob:
        def execute(self):
            return False
        def request_stop(self):
            return True
    callback = lambda _: None
    return Agent(FailedJob(), callback, "failed job")


class JobControl:
    """
    Jobs are pulled out from the left (front of the queue). add_job() appends
    one to the end (right side), while insert_job() inserts it in front (left
    side).
    """
    def __init__(self):
        self._background = {}
        self._active_agent = None
        self._queue = collections.deque()
        self._lock = threading.RLock()

    def clear_queue(self) -> None:
        self._queue.clear()

    def add_job(self, job, name=None):
        return self._enqueue_job(job, self._queue.append, name)

    def insert_job(self, job, name=None):
        return self._enqueue_job(job, self._queue.appendleft, name)

    def spawn_job(self, job, name):
        agent = None
        if self._acquire_lock():
            try:
                agent = Agent(job, self._on_background_done, name)
                self._background[agent.name] = agent
                agent.execute()
            finally:
                self._release_lock()
        return agent

    def get_queued(self):
        return list(self._queue) if self._queue is not None else None

    def get_background(self):
        if self._background is None:
            return None
        return self._background.values()

    def get_current(self):
        return self._active_agent

    def is_running(self, name) -> bool:
        if self._active_agent is not None and self._active_agent.name == name:
            return True
        return name in self._background

    def stop_background(self) -> bool:
        result = False
        if self._acquire_lock():
            result = True
            try:
                # Get a copy to avoid iterating over a list that's undergoing
                # deletions.
                agents = self.get_background()
                if agents is not None:
                    for agent in list(agents).copy():
                        agent.request_stop()
            finally:
                self._release_lock()
        return result

    def stop_job(self, name) -> bool:
        result = False
        if self._acquire_lock():
            try:
                if (self._active_agent is not None and
                        self._active_agent.name == name):
                    self._active_agent.request_stop()
                    result = True
                elif name in self._background:
                    self._background[name].request_stop()
                    result = True
            finally:
                self._release_lock()
        return result

    def stop_current(self) -> bool:
        if self._active_agent is not None and self._active_agent.is_running():
            if self._acquire_lock():
                try:
                    self._active_agent.request_stop()
                finally:
                    self._release_lock()
                return True
        return False

    def has_jobs(self) -> bool:
        return (len(self._queue) > 0 or len(self._background) > 0 or
                self._active_agent is not None)

    def _run_next_job(self) -> None:
        if self._acquire_lock():
            try:
                if self._active_agent is None and len(self._queue) > 0:
                    self._active_agent = self._queue.popleft()
                    self._active_agent.execute()
            finally:
                self._release_lock()

    def _enqueue_job(self, job, append_fn, name):
        agent = None
        if self._acquire_lock():
            try:
                agent = Agent(job, self._on_execution_done, name)
                append_fn(agent)
                if self._active_agent is None:
                    self._run_next_job()
            finally:
                self._lock.release()
        return agent

    def _on_execution_done(self, _):
        if self._acquire_lock():
            try:
                self._active_agent = None
            finally:
                self._release_lock()
            self._run_next_job()

    def _on_background_done(self, agent):
        if self._acquire_lock():
            try:
                del self._background[agent.name]
            finally:
                self._release_lock()

    def _acquire_lock(self):
        if not self._lock.acquire(True, 1.0):
            logging.error("Unable to acquire lock.")
            return False
        return True

    def _release_lock(self):
        self._lock.release()




#!/usr/bin/env python

import argparse
import logging

from bardolph.controller.routine import Routine
if __name__ == '__main__':
    from bardolph.parser.parse import Parser

from .instruction import Instruction
from .vm_codes import JumpCondition, OpCode

class Loader:
    def __init__(self):
        self._main_segment = []
        self._routine_segment = []
        self._iter = None

    def _next_inst(self):
        if self._iter is None:
            return None
        try:
            return next(self._iter)
        except StopIteration:
            self._iter = None
            return None

    def load(self, instructions, routines):
        """ Upon completion routines will be populated as a dictionary
        of Routine objects, keyed on routine name. """
        self._main_segment.clear()
        self._routine_segment.clear()
        if instructions is not None:
            self._iter = iter(instructions)
            inst = self._next_inst()
            while inst is not None:
                if inst.op_code is OpCode.ROUTINE:
                    rtn = self._load_routine(inst)
                    routines[rtn.name] = rtn
                else:
                    self._main_segment.append(inst)
                inst = self._next_inst()

    def _load_routine(self, current_inst):
        self._routine_segment.append(current_inst)
        rtn = Routine(current_inst.param0)
        rtn.set_address(len(self._routine_segment) + 1)

        inst = self._next_inst()
        while inst is not None and inst.op_code is not OpCode.END:
            self._routine_segment.append(inst)
            inst = self._next_inst()
        if inst is not None:
            self._routine_segment.append(inst)
        return rtn

    @property
    def code(self):
        if len(self._routine_segment) == 0:
            return self._main_segment
        ret_value = [Instruction(
            OpCode.JUMP, JumpCondition.ALWAYS, len(self._routine_segment) + 1)]
        ret_value.extend(self._routine_segment)
        ret_value.extend(self._main_segment)
        return ret_value

def main():
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument('file', help='name of the script file')
    args = arg_parser.parse_args()

    logging.basicConfig(
        level=logging.DEBUG,
        format='%(filename)s(%(lineno)d) %(funcName)s(): %(message)s')

    parser = Parser()
    parser_code = parser.load(args.file)

    loader = Loader()
    routines = {}
    loader.load(parser_code, routines)
    if loader.code is not None:
        inst_num = 0
        for inst in loader.code:
            print('{:5d}: {}'.format(inst_num, inst))
            inst_num += 1
    else:
        print("Error parsing: {}".format(parser.get_errors()))


if __name__ == '__main__':
    main()



import logging
from enum import Enum, auto

from bardolph.controller import i_controller
from bardolph.lib.injection import bind_instance


class Action(Enum):
    GET_COLOR = auto()
    GET_ZONE_COLOR = auto()
    GET_POWER = auto()
    SET_COLOR = auto()
    SET_POWER = auto()
    SET_ZONE_COLOR = auto()


class ActivityMonitor:
    def __init__(self):
        self._actions = []
        self._quiet = False

    def log_call(self, name, params=None):
        # params is a tuple or None.
        if not self._quiet:
            self._actions.append((name, params))
        else:
            self._quiet = False

    def log_output(self, msg):
        logging.info(msg)

    def calls_to(self, name):
        # list of tuples and/or None's.
        return [action[1] for action in self._actions if action[0] == name]

    def get_call_list(self):
        return self._actions

    def clear(self):
        self._actions.clear()

    def quietly(self):
        self._quiet = True
        return self


class Light(ActivityMonitor):
    """
    Fake lifxlan.light.Light which implements the methods that are actually
    called by the tests.
    """
    def __init__(self, name, group, location, color=None, multizone=False):
        super().__init__()
        self._name = name
        self._group = group
        self._location = location
        self._multizone = multizone
        self._power = 12345
        self._color = color or [-1] * 4
        self._color_zones = [self._color] * 16
        self._set_color = None
        self._quiet = False

    def __repr__(self):
        fmt = 'fake_lifx.Light(_name: "{}", _group: "{}", _location: "{}", '
        fmt += '_power: {}, _color: {})'
        return fmt.format(
            self._name, self._group, self._location, self._power, self._color)

    def get_color(self):
        self.log_call(Action.GET_COLOR, self._color)
        logging.info(
            'Get color from "{}": {}'.format(self._name, self._color))
        return self._color

    def set_color(self, color, duration=0, _=False):
        self._color = color.copy()
        self._set_color = self._color
        self.log_call(Action.SET_COLOR, (color, duration))
        logging.info(
            'Set color for "{}": {}, {}'.format(
                self._name, color, duration))

    def set_zone_color(self, start_index, end_index, color, duration, _=False):
        for zone in range(start_index, end_index):
            self._color_zones[zone] = color.copy()
        self.log_call(
            Action.SET_ZONE_COLOR, (start_index, end_index, color, duration))
        logging.info('Set color for "{}" zones {} - {}: {}, {}'.format(
            self._name, start_index, end_index, color, duration))

    def supports_multizone(self):
        return self._multizone

    def set_power(self, power, duration, _=False):
        self._power = power
        self.log_call(Action.SET_POWER, (power, duration))
        logging.info(
            'Set power for "{}": {}, {}'.format(self._name, power, duration))

    def get_power(self):
        self.log_call(Action.GET_POWER)
        return self._power

    def get_color_zones(self, start_index=0, end_index=16):
        self.log_call(Action.GET_ZONE_COLOR, (start_index, end_index))
        logging.info('Get color from "{}" zones {} - {}'.format(
                        self._name, start_index, end_index))
        return self._color_zones[start_index : end_index]

    def get_label(self):
        return self._name

    def get_location(self):
        return self._location

    def get_group(self):
        return self._group

    def was_set(self, color):
        return self._set_color == color


class Lifx(i_controller.Lifx, ActivityMonitor):
    _inits = []

    def __init__(self):
        ActivityMonitor.__init__(self)
        self._lights = []
        for init in self._inits:
            self._lights.append(Light(*init))

    def get_lights(self):
        return self._lights

    def set_color_all_lights(self, color, duration):
        self.log_call(Action.SET_COLOR, (color, duration))
        logging.info("Color (all) {}, {}".format(color, duration))
        for light in self.get_lights():
            light.quietly().set_color(color, duration)

    def set_power_all_lights(self, power_level, duration):
        self.log_call(Action.SET_POWER, (power_level, duration))
        logging.info("Power (all) {} {}".format(power_level, duration))
        for light in self.get_lights():
            light.quietly().set_power(power_level, duration)


class _Reinit:
    def configure(self):
        bind_instance(Lifx()).to(i_controller.Lifx)


def configure():
    using_large_set().configure()


def using(inits):
    Lifx._inits = inits.copy()
    return _Reinit()


def using_large_set():
    # light name, group, location
    Lifx._inits = [
        ('Table', 'Furniture', 'Home', [1, 2, 3, 4], False),
        ('Top', 'Pole', 'Home', [10, 20, 30, 40], False),
        ('Middle', 'Pole', 'Home', [100, 200, 300, 400], False),
        ('Bottom', 'Pole', 'Home', [1000, 2000, 3000, 4000], False),
        ('Chair', 'Furniture', 'Home',
            [10000, 20000, 30000, 4004], False),
        ('Strip', 'Furniture', 'Home', [4, 3, 2, 1], True)
    ].copy()
    return _Reinit()


def using_small_set():
    color = [1, 2, 3, 4]
    Lifx._inits = [
        ('light_1', 'x', 'y', color, False),
        ('light_2', 'group', 'loc', color, False),
        ('light_0', 'group', 'loc', color, False)
    ].copy()
    return _Reinit()



from enum import Enum, auto
import logging

class Register(Enum):
    BLUE = auto()
    BRIGHTNESS = auto()
    DURATION = auto()
    FIRST_ZONE = auto()
    GREEN = auto()
    HUE = auto()
    LAST_ZONE = auto()
    KELVIN = auto()
    NAME = auto()
    OPERAND = auto()
    PC = auto()
    POWER = auto()
    RED = auto()
    RESULT = auto()
    SATURATION = auto()
    TIME = auto()
    UNIT_MODE = auto()

    @staticmethod
    def from_string(name):
        upper = name.upper()
        return getattr(Register, upper) if hasattr(Register, upper) else None

class OpCode(Enum):
    BREAKPOINT = auto()
    COLOR = auto()
    CONSTANT = auto()
    DISC = auto()
    DISCM = auto()
    DNEXT = auto()
    DNEXTM = auto()
    END = auto()
    END_LOOP = auto()
    GET_COLOR = auto()
    JSR = auto()
    JUMP = auto()
    LOOP = auto()
    MOVE = auto()
    MOVEQ = auto()
    NOP = auto()
    OP = auto()
    OUT = auto()
    PARAM = auto()
    PAUSE = auto()
    POP = auto()
    POWER = auto()
    PUSH = auto()
    PUSHQ = auto()
    ROUTINE = auto()
    STOP = auto()
    TIME_PATTERN = auto()
    WAIT = auto()

class JumpCondition(Enum):
    ALWAYS = auto()
    IF_FALSE = auto()
    IF_TRUE = auto()
    INDIRECT = auto()

class LoopVar(Enum):
    COUNTER = auto()
    CURRENT = auto()
    EXIT_JMP = auto()
    FIRST = auto()
    INCR = auto()
    LAST = auto()

class Operator(Enum):
    ADD = auto()
    AND = auto()
    DIV = auto()
    EQ = auto()
    LT = auto()
    LTE = auto()
    GT = auto()
    GTE = auto()
    MOD = auto()
    MUL = auto()
    NOT = auto()
    NOTEQ = auto()
    OR = auto()
    POW = auto()
    SUB = auto()
    UADD = auto()
    USUB = auto()

class Operand(Enum):
    ALL = auto()
    LIGHT = auto()
    GROUP = auto()
    LOCATION = auto()
    MZ_LIGHT = auto()
    NULL = auto()

class IoOp(Enum):
    PRINT = auto()
    PRINTF = auto()
    REGISTER = auto()

class SetOp(Enum):
    """ Used with TimePattern """
    INIT = auto()
    UNION = auto()



import copy
import logging

from bardolph.controller import units
from bardolph.controller.get_key import getch
from bardolph.controller.i_controller import LightSet
from bardolph.controller.units import UnitMode
from bardolph.lib.i_lib import Clock, TimePattern
from bardolph.lib.injection import inject, injected, provide
from bardolph.lib.symbol import Symbol

from .call_stack import CallStack
from .loader import Loader
from .vm_codes import JumpCondition, LoopVar, OpCode, Operand, Register, SetOp
from .vm_discover import VmDiscover
from .vm_io import VmIo
from .vm_math import VmMath


class Registers:
    def __init__(self):
        self.blue = 0.0
        self.brightness = 0.0
        self.disc_forward = False
        self.duration = 0.0
        self.first_zone = 0
        self.green = 0.0
        self.hue = 0.0
        self.kelvin = 0.0
        self.last_zone = 0
        self.name = None
        self.operand = Operand.NULL
        self.pc = 0
        self.power = False
        self.red = 0.0
        self.result = None
        self.saturation = 0.0
        self.time = 0.0  # ms.
        self.unit_mode = UnitMode.LOGICAL

    def get_color(self):
        if self.unit_mode is not UnitMode.RGB:
            return [self.hue, self.saturation, self.brightness, self.kelvin]
        return [self.red, self.green, self.blue, self.kelvin]

    def store_color(self, color) -> None:
        if self.unit_mode is not UnitMode.RGB:
            self.hue, self.saturation, self.brightness, self.kelvin = color
        else:
            self.red, self.green, self.blue, self.kelvin = color

    def get_by_enum(self, reg):
        return getattr(self, reg.name.lower())

    def set_by_enum(self, reg, value):
        setattr(self, reg.name.lower(), value)

    def reset(self):
        self.__init__()

    def get_power(self):
        return 65535 if self.power else 0


class MachineState:
    def __init__(self, reg, call_stack):
        self.reg = reg
        self.call_stack = call_stack


class Machine:
    def __init__(self):
        self._cue_time = 0
        self._clock = provide(Clock)
        self._routines = {}
        self._program = []
        self._reg = Registers()
        self._call_stack = CallStack()
        self._vm_io = VmIo(self._call_stack, self._reg)
        self._vm_math = VmMath(self._call_stack, self._reg)
        self._vm_discover = VmDiscover(self._call_stack, self._reg)
        self._enable_pause = True
        self._keep_running = True
        excluded = (OpCode.STOP, OpCode.ROUTINE)
        op_codes = [code for code in OpCode
                    if not str(code.name).startswith('_')
                    and code not in excluded]
        self._fn_table = {
            op_code: getattr(self, '_' + op_code.name.lower())
            for op_code in (op_codes)}
        self._fn_table[OpCode.STOP] = self.stop

    def reset(self) -> None:
        self._reg.reset()
        self._routines.clear()
        self._cue_time = 0
        self._call_stack.reset()
        self._vm_math.reset()
        self._keep_running = True
        self._enable_pause = True

    def run(self, program) -> None:
        loader = Loader()
        loader.load(program, self._routines)
        self._program = loader.code
        self._keep_running = True

        logging.debug('Starting to execute.')
        self._clock.start()
        program_len = len(self._program)
        try:
            while self._keep_running and self._reg.pc < program_len:
                inst = self._program[self._reg.pc]
                if inst.op_code == OpCode.STOP:
                    break
                self._fn_table[inst.op_code]()
                if inst.op_code not in (OpCode.END, OpCode.JSR, OpCode.JUMP):
                    self._reg.pc += 1
            self._clock.stop()
            logging.debug(
                'Stopped, _keep_running = {}, _pc = {}, program_len = {}'
                .format(
                    self._keep_running, self._reg.pc, program_len))
        except Exception as ex:
            logging.error("Machine stopped due to {}".format(ex))

    def stop(self) -> None:
        self._keep_running = False
        self._clock.stop()

    def get_state(self) -> MachineState:
        return MachineState(self._reg, self._call_stack)

    def color_to_reg(self, color) -> None:
        reg = self._reg
        if reg.unit_mode in (UnitMode.RAW, UnitMode.LOGICAL):
            reg.hue, reg.saturation, reg.brightness, reg.kelvin = color
        else:
            reg.red, reg.green, reg.blue, reg.kelvin = color

    def color_from_reg(self):
        return self._reg.get_color()

    def get_variable(self, name):
        return self._call_stack.get_variable(name)

    @property
    def current_inst(self):
        return self._program[self._reg.pc]

    def _color(self) -> None:
        fn_map = {operand: fn for operand, fn in (
            (Operand.ALL, self._color_all),
            (Operand.LIGHT, self._color_light),
            (Operand.GROUP, self._color_group),
            (Operand.LOCATION, self._color_location),
            (Operand.MZ_LIGHT, self._color_mz_light))}
        fn_map[self._reg.operand]()

    @inject(LightSet)
    def _color_all(self, light_set=injected) -> None:
        color = self._assure_raw_color(self._reg.get_color())
        duration = self._assure_raw_time(self._reg.duration)
        light_set.set_color(color, duration)

    @inject(LightSet)
    def _color_light(self, light_set=injected) -> None:
        light = light_set.get_light(self._reg.name)
        if light is None:
            Machine._report_missing(self._reg.name)
        else:
            light.set_color(
                self._assure_raw_color(self._reg.get_color()),
                self._assure_raw_time(self._reg.duration))

    @inject(LightSet)
    def _color_mz_light(self, light_set=injected) -> None:
        light = light_set.get_light(self._reg.name)
        if light is None:
            Machine._report_missing(self._reg.name)
        elif self._zone_check(light):
            # Unknown why this happens.
            if not hasattr(light, 'set_zone_color'):
                logging.error(
                    'No set_zone_color for light of type', type(light))
            else:
                start_index = self._reg.first_zone
                end_index = self._reg.last_zone
                if end_index is None:
                    end_index = start_index
                light.set_zone_color(
                    start_index, end_index + 1,
                    self._assure_raw_color(self._reg.get_color()),
                    self._assure_raw_time(self._reg.duration))

    @inject(LightSet)
    def _color_group(self, light_set=injected) -> None:
        light_names = light_set.get_group(self._reg.name)
        if light_names is None:
            logging.warning("Unknown group: {}".format(self._reg.name))
        else:
            self._color_multiple(
                [light_set.get_light(name) for name in light_names])

    @inject(LightSet)
    def _color_location(self, light_set=injected) -> None:
        light_names = light_set.get_location(self._reg.name)
        if light_names is None:
            logging.warning("Unknown location: {}".format(self._reg.name))
        else:
            self._color_multiple(
                [light_set.get_light(name) for name in light_names])

    def _color_multiple(self, lights) -> None:
        color = self._assure_raw_color(self._reg.get_color())
        duration = self._assure_raw_time(self._reg.duration)
        for light in lights:
            light.set_color(color, duration)

    def _power(self) -> None: {
            Operand.ALL: self._power_all,
            Operand.LIGHT: self._power_light,
            Operand.GROUP: self._power_group,
            Operand.LOCATION: self._power_location
        }[self._reg.operand]()

    @inject(LightSet)
    def _power_all(self, light_set=injected) -> None:
        duration = self._assure_raw_time(self._reg.duration)
        light_set.set_power(self._reg.get_power(), duration)

    @inject(LightSet)
    def _power_light(self, light_set=injected) -> None:
        light = light_set.get_light(self._reg.name)
        if light is None:
            Machine._report_missing(self._reg.name)
        else:
            duration = self._assure_raw_time(self._reg.duration)
            light.set_power(self._reg.get_power(), duration)

    @inject(LightSet)
    def _power_group(self, light_set=injected) -> None:
        light_names = light_set.get_group(self._reg.name)
        if light_names is None:
            logging.warning(
                'Power invoked for unknown group "{}"'.format(self._reg.name))
        else:
            self._power_multiple(
                [light_set.get_light(name) for name in light_names])

    @inject(LightSet)
    def _power_location(self, light_set=injected) -> None:
        light_names = light_set.get_location(self._reg.name)
        if light_names is None:
            logging.warning(
                "Power invoked for unknown location: {}".format(self._reg.name))
        else:
            self._power_multiple(
                [light_set.get_light(name) for name in light_names])

    def _power_multiple(self, lights) -> None:
        power = self._reg.get_power()
        for light in lights:
            light.set_power(power, self._reg.duration)

    @inject(LightSet)
    def _get_color(self, light_set=injected) -> None:
        light = light_set.get_light(self._reg.name)
        if light is None:
            Machine._report_missing(self._reg.name)
        else:
            if self._reg.operand is Operand.MZ_LIGHT:
                if self._zone_check(light):
                    zone = self._reg.first_zone
                    color = light.get_color_zones(zone, zone + 1)[0]
                    self.color_to_reg(self._maybe_converted_color(color))
            else:
                color = light.get_color()
                self.color_to_reg(self._maybe_converted_color(color))

    def _param(self) -> None:
        """
        param instruction: the name of the parameter is in param0, and its
        value is in param1. If the value is a Symbol or Register, it needs to
        be dereferenced.
        """
        inst = self.current_inst
        value = inst.param1
        if isinstance(value, Symbol):
            value = self._call_stack.get_variable(value.name)
        elif isinstance(value, Register):
            value = self._reg.get_by_enum(value)
        self._call_stack.put_param(inst.param0, value)

    def _jsr(self) -> None:
        inst = self.current_inst
        self._call_stack.set_return(self._reg.pc + 1)
        self._call_stack.push_current()
        routine_name = inst.param0
        rtn = self._routines.get(routine_name, None)
        self._reg.pc = rtn.get_address()

    def _end(self) -> None:
        self._call_stack.unwind_loops()
        ret_addr = self._call_stack.get_return()
        self._call_stack.pop_current()
        self._reg.pc = ret_addr

    def _jump(self) -> None:
        # In the current instruction, param0 contains the condition, and
        # param1 contains the offset.
        inst = self.current_inst
        if inst.param0 is JumpCondition.INDIRECT:
            address = self._call_stack.get_variable(inst.param1)
            self._reg.pc = address
        else:
            jump_if = {
                JumpCondition.ALWAYS: {True: True, False: True},
                JumpCondition.IF_FALSE: {True: False, False: True},
                JumpCondition.IF_TRUE: {True: True, False: False}
            }
            if jump_if[inst.param0][bool(self._reg.result)]:
                self._reg.pc += inst.param1
            else:
                self._reg.pc += 1

    def _loop(self) -> None:
        self._call_stack.enter_loop()

    def _end_loop(self) -> None:
        self._call_stack.exit_loop()

    def _nop(self) -> None: pass

    def _push(self) -> None:
        self._vm_math.push(self.current_inst.param0)

    def _pushq(self) -> None:
        self._vm_math.pushq(self.current_inst.param0)

    def _pop(self) -> None:
        self._vm_math.pop(self.current_inst.param0)

    def _op(self) -> None:
        self._vm_math.op(self.current_inst.param0)

    def _bin_op(self, operator) -> None:
        try:
            self._vm_math.bin_op(operator)
        except ZeroDivisionError:
            self._trigger_error("Division by zero. Halting execution.")
            self.stop()

    def _unary_op(self, operator) -> None:
        self._vm_math.unary_op(operator)

    def _disc(self) -> None:
        self._vm_discover.disc()

    def _discm(self) -> None:
            self._vm_discover.discm(self.current_inst.param0)

    def _dnext(self) -> None:
        self._vm_discover.dnext(self.current_inst.param0)

    def _dnextm(self) -> None:
        self._vm_discover.dnextm(
            self.current_inst.param0, self.current_inst.param1)

    def _out(self) -> None:
        self._vm_io.out(self.current_inst)

    def _pause(self) -> None:
        if self._enable_pause:
            print("Press any key to continue, q to quit, "
                  + "! to run without stopping again.")
            char = getch()
            if char == 'q':
                self.stop()
            else:
                print("Running...")
                if char == '!':
                    self._enable_pause = False

    def _constant(self) -> None:
        name = self.current_inst.param0
        value = self.current_inst.param1
        self._call_stack.put_constant(name, value)

    def _wait(self) -> None:
        time = self._reg.time
        if isinstance(time, TimePattern):
            self._clock.wait_until(time)
        elif time > 0:
            if self._reg.unit_mode is UnitMode.RAW:
                time /= 1000.0
            self._clock.pause_for(time)

    def _assure_raw_time(self, value) -> int:
        if self._reg.unit_mode in (UnitMode.LOGICAL, UnitMode.RGB):
            return units.time_raw(value)
        return value

    def _assure_raw_color(self, color):
        if self._reg.unit_mode is UnitMode.RAW:
            return color
        if self._reg.unit_mode is UnitMode.RGB:
            return units.rgb_to_raw(color)
        return units.logical_to_raw(color)

    def _maybe_converted_color(self, color):
        """
        The incoming color always consists of raw values.
        """
        if self._reg.unit_mode is UnitMode.RAW:
            return color
        if self._reg.unit_mode is UnitMode.LOGICAL:
            return units.raw_to_logical(color)
        return units.raw_to_rgb(color)

    def _move(self) -> None:
        """
        Move from variable/register to variable/register.
        """
        inst = self.current_inst
        value = inst.param0
        dest = inst.param1
        if isinstance(value, Register):
            value = self._reg.get_by_enum(value)
        elif isinstance(value, (str, LoopVar)):
            value = self._call_stack.get_variable(value)
        self._do_put_value(dest, value)

    def _moveq(self) -> None:
        """
        Move a value from the instruction itself into a register or variable.
        """
        value = self.current_inst.param0
        dest = self.current_inst.param1
        if dest is Register.UNIT_MODE:
            self._switch_unit_mode(value)
        else:
            self._do_put_value(dest, value)

    @staticmethod
    def _convert_units_fn(from_mode, to_mode):
        def key(mode0, mode1): return str(mode0) + str(mode1)
        converters = (
            (UnitMode.LOGICAL, UnitMode.RAW, units.logical_to_raw),
            (UnitMode.LOGICAL, UnitMode.RGB, units.logical_to_rgb),
            (UnitMode.RGB, UnitMode.RAW, units.rgb_to_raw),
            (UnitMode.RGB, UnitMode.LOGICAL, units.rgb_to_logical),
            (UnitMode.RAW, UnitMode.RGB, units.raw_to_rgb),
            (UnitMode.RAW, UnitMode.LOGICAL, units.raw_to_logical))
        convert_dict = {key(from_mode, to_mode): fn
                        for from_mode, to_mode, fn in converters}
        return convert_dict[key(from_mode, to_mode)]

    def _switch_unit_mode(self, to_mode) -> None:
        from_mode = self._reg.unit_mode
        if from_mode is to_mode:
            return

        original_color = self._reg.get_color()
        self._reg.unit_mode = to_mode
        converter = self._convert_units_fn(from_mode, to_mode)
        self._reg.store_color(converter(original_color))

        if to_mode is UnitMode.RAW:
            self._reg.duration = units.time_raw(self._reg.duration)
            self._reg.time = units.time_raw(self._reg.time)
        elif from_mode is UnitMode.RAW:
            self._reg.duration = units.time_logical(self._reg.duration)
            self._reg.time = units.time_logical(self._reg.time)

    def _do_put_value(self, dest, value) -> None:
        if isinstance(dest, Register):
            self._reg.set_by_enum(dest, value)
        else:
            self._call_stack.put_variable(dest, value)

    def _time_pattern(self) -> None:
        inst = self.current_inst
        if inst.param0 == SetOp.INIT:
            self._reg.time = inst.param1
        else:
            self._reg.time.union(inst.param1)

    def _zone_check(self, light) -> bool:
        if not light.multizone:
            logging.warning(
                'Light "{}" is not multi-zone.'.format(light.name))
            return False
        return True

    @staticmethod
    def _report_missing(name) -> None:
        logging.warning('Light "{}" not found.'.format(name))

    def _power_param(self):
        return 65535 if self._reg.power else 0

    def _breakpoint(self) -> None:
        print("At breakpoint.")

    def _trigger_error(self, message) -> bool:
        logging.error(message)
        return False



import sys
import re

from bardolph.lib.time_pattern import TimePattern

from .token import Token, TokenTypes


class Lex:
    _CMP_SPEC = r'==|<=|>=|!=|[<>]'
    _REG = 'hue saturation brightness kelvin red green blue duration time'
    _REG_LIST = _REG.split()
    _NAME_SPEC = r'[a-zA-Z_][a-zA-Z0-9_]*'
    _NON_ALNUM_LIST = r'[]{}()+-*/#:^'
    _NON_ALNUM_SPEC = r'==|<=|>=|[\[\]\(\){}+\-*<>/#:\^]'
    _NUMBER_SPEC = r'[0-9]*\.?[0-9]+'
    _LITERAL_STRING_SPEC = r'"([^"]|(?<=\\)")*"'
    _DEFAULT_SPEC = '[^\s]+'

    _STRING = re.compile(_LITERAL_STRING_SPEC)
    _TOKEN_SPEC = '|'.join((
        TimePattern.REGEX_SPEC,
        _CMP_SPEC,
        _LITERAL_STRING_SPEC,
        _NUMBER_SPEC,
        _NAME_SPEC,
        _NON_ALNUM_SPEC,
        _DEFAULT_SPEC))

    _CMP = re.compile(_CMP_SPEC)
    _TOKEN = re.compile(_TOKEN_SPEC)
    _NAME = re.compile(_NAME_SPEC)
    _NUMBER = re.compile(_NUMBER_SPEC)
    _TIME_PATTERN = TimePattern.REGEX
    _INT = re.compile(r'^\-?[0-9]*$')

    def __init__(self, input_string, source=''):
        self._lines = iter(input_string.split('\n'))
        self._source = source

    def tokens(self):
        line_num = 0
        for line in self._lines:
            line_num += 1
            for match in self._TOKEN.finditer(line):
                matched = match.string[match.start():match.end()]
                u_matched = self._unabbreviate(matched)
                if u_matched == '#':
                    break
                if u_matched in self._NON_ALNUM_LIST:
                    token_type = TokenTypes.MARK
                else:
                    token_type = self._token_type(u_matched)
                    if token_type is TokenTypes.LITERAL_STRING:
                        u_matched = u_matched[1:-1]
                        u_matched = u_matched.replace(r'\"', '"')
                yield Token(token_type, u_matched, line_num, self._source)
        yield Token(TokenTypes.EOF)

    @staticmethod
    def is_int(text):
        return Lex._INT.match(text) is not None

    def _token_type(self, word):
        token_type = TokenTypes.__members__.get(word.upper())
        if token_type is not None:
            return token_type
        if word in self._REG_LIST:
            return TokenTypes.REGISTER
        pairs = (
            (self._CMP, TokenTypes.COMPARE),
            (self._TIME_PATTERN, TokenTypes.TIME_PATTERN),
            (self._STRING, TokenTypes.LITERAL_STRING),
            (self._NUMBER, TokenTypes.NUMBER),
            (self._NAME, TokenTypes.NAME))
        for reg_expr, token_type in pairs:
            if reg_expr.match(word):
                return token_type
        return TokenTypes.ERROR

    @staticmethod
    def _unabbreviate(token):
        return {
            'H': 'hue', 'S': 'saturation', 'B': 'brightness', 'K': 'kelvin'
        }.get(token, token)


def main():
    args = sys.argv[1:]
    args = ['(abc)']
    for arg in args:
        print("lexing: {}\n".format(arg))
        lexer = Lex(arg)
        for token in lexer.tokens():
            print("token: {}\ntype: {}\n".format(
                str(token), token.token_type))


if __name__ == '__main__':
    main()


def average_color(colors):
    """
    Returns average for the list, or None of the list is empty. Each element is
    an inner list of 4 numbers.
    """
    count = len(colors)
    if count == 0:
        return None
    totals = [0, 0, 0, 0]
    for color in colors:
        for i in range(0, 4):
            totals[i] += color[i]
    avg = []
    for i in range(0, 4):
        avg.append(round(totals[i] / count))
    return avg

def rounded_color(color):
    return [round(c) for c in color]
